import io
from test._async_compat import mark_async_test
from unittest.mock import patch

import pytest
from neo4j.exceptions import ClientError

from neomodel import (
    ArrayProperty,
    AsyncRelationshipTo,
    AsyncStructuredNode,
    AsyncStructuredRel,
    FloatProperty,
    FulltextIndex,
    StringProperty,
    UniqueIdProperty,
    VectorIndex,
    adb,
)
from neomodel.exceptions import ConstraintValidationFailed, FeatureNotSupported


class NodeWithIndex(AsyncStructuredNode):
    name = StringProperty(index=True)


class NodeWithConstraint(AsyncStructuredNode):
    name = StringProperty(unique_index=True)


class NodeWithRelationship(AsyncStructuredNode):
    ...


class IndexedRelationship(AsyncStructuredRel):
    indexed_rel_prop = StringProperty(index=True)


class OtherNodeWithRelationship(AsyncStructuredNode):
    has_rel = AsyncRelationshipTo(
        NodeWithRelationship, "INDEXED_REL", model=IndexedRelationship
    )


class AbstractNode(AsyncStructuredNode):
    __abstract_node__ = True
    name = StringProperty(unique_index=True)


class SomeNotUniqueNode(AsyncStructuredNode):
    id_ = UniqueIdProperty(db_property="id")


@mark_async_test
async def test_install_all():
    await adb.drop_constraints()
    await adb.drop_indexes()
    await adb.install_labels(AbstractNode)
    # run install all labels
    await adb.install_all_labels()

    indexes = await adb.list_indexes()
    index_names = [index["name"] for index in indexes]
    assert "index_INDEXED_REL_indexed_rel_prop" in index_names

    constraints = await adb.list_constraints()
    constraint_names = [constraint["name"] for constraint in constraints]
    assert "constraint_unique_NodeWithConstraint_name" in constraint_names
    assert "constraint_unique_SomeNotUniqueNode_id" in constraint_names

    # remove constraint for above test
    await _drop_constraints_for_label_and_property("NoConstraintsSetup", "name")


@mark_async_test
async def test_install_label_twice(capsys):
    await adb.drop_constraints()
    await adb.drop_indexes()
    expected_std_out = (
        "{code: Neo.ClientError.Schema.EquivalentSchemaRuleAlreadyExists}"
    )
    await adb.install_labels(AbstractNode)
    await adb.install_labels(AbstractNode)

    await adb.install_labels(NodeWithIndex)
    await adb.install_labels(NodeWithIndex, quiet=False)
    captured = capsys.readouterr()
    assert expected_std_out in captured.out

    await adb.install_labels(NodeWithConstraint)
    await adb.install_labels(NodeWithConstraint, quiet=False)
    captured = capsys.readouterr()
    assert expected_std_out in captured.out

    await adb.install_labels(OtherNodeWithRelationship)
    await adb.install_labels(OtherNodeWithRelationship, quiet=False)
    captured = capsys.readouterr()
    assert expected_std_out in captured.out

    if await adb.version_is_higher_than("5.7"):

        class UniqueIndexRelationship(AsyncStructuredRel):
            unique_index_rel_prop = StringProperty(unique_index=True)

        class OtherNodeWithUniqueIndexRelationship(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                NodeWithRelationship, "UNIQUE_INDEX_REL", model=UniqueIndexRelationship
            )

        await adb.install_labels(OtherNodeWithUniqueIndexRelationship)
        await adb.install_labels(OtherNodeWithUniqueIndexRelationship, quiet=False)
        captured = capsys.readouterr()
        assert expected_std_out in captured.out


@mark_async_test
async def test_install_labels_db_property(capsys):
    await adb.drop_constraints()
    await adb.install_labels(SomeNotUniqueNode, quiet=False)
    captured = capsys.readouterr()
    assert "id" in captured.out
    # make sure that the id_ constraint doesn't exist
    constraint_names = await _drop_constraints_for_label_and_property(
        "SomeNotUniqueNode", "id_"
    )
    assert constraint_names == []
    # make sure the id constraint exists and can be removed
    await _drop_constraints_for_label_and_property("SomeNotUniqueNode", "id")


@mark_async_test
async def test_relationship_unique_index_not_supported():
    if await adb.version_is_higher_than("5.7"):
        pytest.skip("Not supported before 5.7")

    class UniqueIndexRelationship(AsyncStructuredRel):
        name = StringProperty(unique_index=True)

    class TargetNodeForUniqueIndexRelationship(AsyncStructuredNode):
        pass

    with pytest.raises(
        FeatureNotSupported, match=r".*Please upgrade to Neo4j 5.7 or higher"
    ):

        class NodeWithUniqueIndexRelationship(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                TargetNodeForUniqueIndexRelationship,
                "UNIQUE_INDEX_REL",
                model=UniqueIndexRelationship,
            )

        await adb.install_labels(NodeWithUniqueIndexRelationship)


@mark_async_test
async def test_relationship_unique_index():
    if not await adb.version_is_higher_than("5.7"):
        pytest.skip("Not supported before 5.7")

    class UniqueIndexRelationshipBis(AsyncStructuredRel):
        name = StringProperty(unique_index=True)

    class TargetNodeForUniqueIndexRelationship(AsyncStructuredNode):
        pass

    class NodeWithUniqueIndexRelationship(AsyncStructuredNode):
        has_rel = AsyncRelationshipTo(
            TargetNodeForUniqueIndexRelationship,
            "UNIQUE_INDEX_REL_BIS",
            model=UniqueIndexRelationshipBis,
        )

    await adb.install_labels(NodeWithUniqueIndexRelationship)
    node1 = await NodeWithUniqueIndexRelationship().save()
    node2 = await TargetNodeForUniqueIndexRelationship().save()
    node3 = await TargetNodeForUniqueIndexRelationship().save()
    rel1 = await node1.has_rel.connect(node2, {"name": "rel1"})

    with pytest.raises(
        ConstraintValidationFailed,
        match=r".*already exists with type `UNIQUE_INDEX_REL_BIS` and property `name`.*",
    ):
        rel2 = await node1.has_rel.connect(node3, {"name": "rel1"})


@mark_async_test
async def test_fulltext_index():
    if not await adb.version_is_higher_than("5.16"):
        pytest.skip("Not supported before 5.16")

    class FullTextIndexNode(AsyncStructuredNode):
        name = StringProperty(fulltext_index=FulltextIndex(eventually_consistent=True))

    await adb.install_labels(FullTextIndexNode)
    indexes = await adb.list_indexes()
    index_names = [index["name"] for index in indexes]
    assert "fulltext_index_FullTextIndexNode_name" in index_names

    await adb.cypher_query("DROP INDEX fulltext_index_FullTextIndexNode_name")


@mark_async_test
async def test_fulltext_index_conflict():
    if not await adb.version_is_higher_than("5.16"):
        pytest.skip("Not supported before 5.16")

    stream = io.StringIO()

    with patch("sys.stdout", new=stream):
        await adb.cypher_query(
            "CREATE FULLTEXT INDEX FOR (n:FullTextIndexNodeConflict) ON EACH [n.name]"
        )

        class FullTextIndexNodeConflict(AsyncStructuredNode):
            name = StringProperty(fulltext_index=FulltextIndex())

        await adb.install_labels(FullTextIndexNodeConflict, quiet=False)

    console_output = stream.getvalue()
    assert "Creating fulltext index" in console_output
    assert "There already exists an index" in console_output


@mark_async_test
async def test_fulltext_index_not_supported():
    if await adb.version_is_higher_than("5.16"):
        pytest.skip("Test only for versions lower than 5.16")

    with pytest.raises(
        FeatureNotSupported, match=r".*Please upgrade to Neo4j 5.16 or higher"
    ):

        class FullTextIndexNodeOld(AsyncStructuredNode):
            name = StringProperty(fulltext_index=FulltextIndex())

        await adb.install_labels(FullTextIndexNodeOld)


@mark_async_test
async def test_rel_fulltext_index():
    if not await adb.version_is_higher_than("5.16"):
        pytest.skip("Not supported before 5.16")

    class FullTextIndexRel(AsyncStructuredRel):
        name = StringProperty(fulltext_index=FulltextIndex(eventually_consistent=True))

    class FullTextIndexRelNode(AsyncStructuredNode):
        has_rel = AsyncRelationshipTo(
            "FullTextIndexRelNode", "FULLTEXT_INDEX_REL", model=FullTextIndexRel
        )

    await adb.install_labels(FullTextIndexRelNode)
    indexes = await adb.list_indexes()
    index_names = [index["name"] for index in indexes]
    assert "fulltext_index_FULLTEXT_INDEX_REL_name" in index_names

    await adb.cypher_query("DROP INDEX fulltext_index_FULLTEXT_INDEX_REL_name")


@mark_async_test
async def test_rel_fulltext_index_conflict():
    if not await adb.version_is_higher_than("5.16"):
        pytest.skip("Not supported before 5.16")

    stream = io.StringIO()

    with patch("sys.stdout", new=stream):
        await adb.cypher_query(
            "CREATE FULLTEXT INDEX FOR ()-[r:FULLTEXT_INDEX_REL_CONFLICT]-() ON EACH [r.name]"
        )

        class FullTextIndexRelConflict(AsyncStructuredRel):
            name = StringProperty(
                fulltext_index=FulltextIndex(eventually_consistent=True)
            )

        class FullTextIndexRelConflictNode(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                "FullTextIndexRelConflictNode",
                "FULLTEXT_INDEX_REL_CONFLICT",
                model=FullTextIndexRelConflict,
            )

        await adb.install_labels(FullTextIndexRelConflictNode, quiet=False)

    console_output = stream.getvalue()
    assert "Creating fulltext index" in console_output
    assert "There already exists an index" in console_output


@mark_async_test
async def test_rel_fulltext_index_not_supported():
    if await adb.version_is_higher_than("5.16"):
        pytest.skip("Test only for versions lower than 5.16")

    with pytest.raises(
        FeatureNotSupported, match=r".*Please upgrade to Neo4j 5.16 or higher"
    ):

        class FullTextIndexRelOld(AsyncStructuredRel):
            name = StringProperty(
                fulltext_index=FulltextIndex(eventually_consistent=True)
            )

        class FullTextIndexRelOldNode(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                "FullTextIndexRelOldNode",
                "FULLTEXT_INDEX_REL_OLD",
                model=FullTextIndexRelOld,
            )

        await adb.install_labels(FullTextIndexRelOldNode)


@mark_async_test
async def test_vector_index():
    if not await adb.version_is_higher_than("5.15"):
        pytest.skip("Not supported before 5.15")

    class VectorIndexNode(AsyncStructuredNode):
        embedding = ArrayProperty(
            FloatProperty(),
            vector_index=VectorIndex(dimensions=256, similarity_function="euclidean"),
        )

    await adb.install_labels(VectorIndexNode)
    indexes = await adb.list_indexes()
    index_names = [index["name"] for index in indexes]
    assert "vector_index_VectorIndexNode_embedding" in index_names

    await adb.cypher_query("DROP INDEX vector_index_VectorIndexNode_embedding")


@mark_async_test
async def test_vector_index_conflict():
    if not await adb.version_is_higher_than("5.15"):
        pytest.skip("Not supported before 5.15")

    stream = io.StringIO()

    with patch("sys.stdout", new=stream):
        await adb.cypher_query(
            "CREATE VECTOR INDEX FOR (n:VectorIndexNodeConflict) ON n.embedding OPTIONS{indexConfig:{`vector.similarity_function`:'cosine', `vector.dimensions`:1536}}"
        )

        class VectorIndexNodeConflict(AsyncStructuredNode):
            embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

        await adb.install_labels(VectorIndexNodeConflict, quiet=False)

    console_output = stream.getvalue()
    assert "Creating vector index" in console_output
    assert "There already exists an index" in console_output


@mark_async_test
async def test_vector_index_not_supported():
    if await adb.version_is_higher_than("5.15"):
        pytest.skip("Test only for versions lower than 5.15")

    with pytest.raises(
        FeatureNotSupported, match=r".*Please upgrade to Neo4j 5.15 or higher"
    ):

        class VectorIndexNodeOld(AsyncStructuredNode):
            embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

        await adb.install_labels(VectorIndexNodeOld)


@mark_async_test
async def test_rel_vector_index():
    if not await adb.version_is_higher_than("5.18"):
        pytest.skip("Not supported before 5.18")

    class VectorIndexRel(AsyncStructuredRel):
        embedding = ArrayProperty(
            FloatProperty(),
            vector_index=VectorIndex(dimensions=256, similarity_function="euclidean"),
        )

    class VectorIndexRelNode(AsyncStructuredNode):
        has_rel = AsyncRelationshipTo(
            "VectorIndexRelNode", "VECTOR_INDEX_REL", model=VectorIndexRel
        )

    await adb.install_labels(VectorIndexRelNode)
    indexes = await adb.list_indexes()
    index_names = [index["name"] for index in indexes]
    assert "vector_index_VECTOR_INDEX_REL_embedding" in index_names

    await adb.cypher_query("DROP INDEX vector_index_VECTOR_INDEX_REL_embedding")


@mark_async_test
async def test_rel_vector_index_conflict():
    if not await adb.version_is_higher_than("5.18"):
        pytest.skip("Not supported before 5.18")

    stream = io.StringIO()

    with patch("sys.stdout", new=stream):
        await adb.cypher_query(
            "CREATE VECTOR INDEX FOR ()-[r:VECTOR_INDEX_REL_CONFLICT]-() ON r.embedding OPTIONS{indexConfig:{`vector.similarity_function`:'cosine', `vector.dimensions`:1536}}"
        )

        class VectorIndexRelConflict(AsyncStructuredRel):
            embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

        class VectorIndexRelConflictNode(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                "VectorIndexRelConflictNode",
                "VECTOR_INDEX_REL_CONFLICT",
                model=VectorIndexRelConflict,
            )

        await adb.install_labels(VectorIndexRelConflictNode, quiet=False)

    console_output = stream.getvalue()
    assert "Creating vector index" in console_output
    assert "There already exists an index" in console_output


@mark_async_test
async def test_rel_vector_index_not_supported():
    if await adb.version_is_higher_than("5.18"):
        pytest.skip("Test only for versions lower than 5.18")

    with pytest.raises(
        FeatureNotSupported, match=r".*Please upgrade to Neo4j 5.18 or higher"
    ):

        class VectorIndexRelOld(AsyncStructuredRel):
            embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

        class VectorIndexRelOldNode(AsyncStructuredNode):
            has_rel = AsyncRelationshipTo(
                "VectorIndexRelOldNode",
                "VECTOR_INDEX_REL_OLD",
                model=VectorIndexRelOld,
            )

        await adb.install_labels(VectorIndexRelOldNode)


@mark_async_test
async def test_unauthorized_index_creation():
    if not await adb.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")

    unauthorized_user = "troygreene"
    expected_message_index = r".*Schema operation.* not allowed for user.*"

    # Standard node index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedIndexNode(AsyncStructuredNode):
                name = StringProperty(index=True)

            await adb.install_labels(UnauthorizedIndexNode)

    # Node uniqueness constraint
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedUniqueIndexNode(AsyncStructuredNode):
                name = StringProperty(unique_index=True)

            await adb.install_labels(UnauthorizedUniqueIndexNode)

    # Relationship index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedRelIndex(AsyncStructuredRel):
                name = StringProperty(index=True)

            class UnauthorizedRelIndexNode(AsyncStructuredNode):
                has_rel = AsyncRelationshipTo(
                    "UnauthorizedRelIndexNode",
                    "UNAUTHORIZED_REL_INDEX",
                    model=UnauthorizedRelIndex,
                )

            await adb.install_labels(UnauthorizedRelIndexNode)


@mark_async_test
async def test_unauthorized_index_creation_recent_features():
    if not await adb.edition_is_enterprise() or not await adb.version_is_higher_than(
        "5.18"
    ):
        pytest.skip("Skipping test for community edition and versions lower than 5.18")

    unauthorized_user = "troygreene"
    expected_message_index = r".*Schema operation.* not allowed for user.*"

    # Node fulltext index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedFulltextNode(AsyncStructuredNode):
                name = StringProperty(fulltext_index=FulltextIndex())

            await adb.install_labels(UnauthorizedFulltextNode)

    # Node vector index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedVectorNode(AsyncStructuredNode):
                embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

            await adb.install_labels(UnauthorizedVectorNode)

    # Relationship uniqueness constraint
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedUniqueRel(AsyncStructuredRel):
                name = StringProperty(unique_index=True)

            class UnauthorizedUniqueRelNode(AsyncStructuredNode):
                has_rel = AsyncRelationshipTo(
                    "UnauthorizedUniqueRelNode",
                    "UNAUTHORIZED_UNIQUE_REL",
                    model=UnauthorizedUniqueRel,
                )

            await adb.install_labels(UnauthorizedUniqueRelNode)

    # Relationship fulltext index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedFulltextRel(AsyncStructuredRel):
                name = StringProperty(fulltext_index=FulltextIndex())

            class UnauthorizedFulltextRelNode(AsyncStructuredNode):
                has_rel = AsyncRelationshipTo(
                    "UnauthorizedFulltextRelNode",
                    "UNAUTHORIZED_FULLTEXT_REL",
                    model=UnauthorizedFulltextRel,
                )

            await adb.install_labels(UnauthorizedFulltextRelNode)

    # Relationship vector index
    with pytest.raises(
        ClientError,
        match=expected_message_index,
    ):
        with await adb.impersonate(unauthorized_user):

            class UnauthorizedVectorRel(AsyncStructuredRel):
                embedding = ArrayProperty(FloatProperty(), vector_index=VectorIndex())

            class UnauthorizedVectorRelNode(AsyncStructuredNode):
                has_rel = AsyncRelationshipTo(
                    "UnauthorizedVectorRelNode",
                    "UNAUTHORIZED_VECTOR_REL",
                    model=UnauthorizedVectorRel,
                )

            await adb.install_labels(UnauthorizedVectorRelNode)


async def _drop_constraints_for_label_and_property(
    label: str = None, property: str = None
):
    results, meta = await adb.cypher_query("SHOW CONSTRAINTS")
    results_as_dict = [dict(zip(meta, row)) for row in results]
    constraint_names = [
        constraint
        for constraint in results_as_dict
        if constraint["labelsOrTypes"] == label and constraint["properties"] == property
    ]
    for constraint_name in constraint_names:
        await adb.cypher_query(f"DROP CONSTRAINT {constraint_name}")

    return constraint_names



"""
Copyright (c) Django Software Foundation and individual contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of Django nor the names of its contributors may be used
       to endorse or promote products derived from this software without
       specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


A class for storing a tree graph. Primarily used for filter constructs in the
ORM.
"""

import copy


class QBase:
    """
    A single internal node in the tree graph. A Node should be viewed as a
    connection (the root) with the children being either leaf nodes or other
    Node instances.
    """

    # Standard connector type. Clients usually won't use this at all and
    # subclasses will usually override the value.
    default = "DEFAULT"

    def __init__(self, children=None, connector=None, negated=False):
        """Construct a new Node. If no connector is given, use the default."""
        self.children = children[:] if children else []
        self.connector = connector or self.default
        self.negated = negated

    # Required because django.db.models.query_utils.Q. Q. __init__() is
    # problematic, but it is a natural Node subclass in all other respects.
    @classmethod
    def _new_instance(cls, children=None, connector=None, negated=False):
        """
        Create a new instance of this class when new Nodes (or subclasses) are
        needed in the internal code in this class. Normally, it just shadows
        __init__(). However, subclasses with an __init__ signature that aren't
        an extension of Node.__init__ might need to implement this method to
        allow a Node to create a new instance of them (if they have any extra
        setting up to do).
        """
        obj = QBase(children, connector, negated)
        obj.__class__ = cls
        return obj

    def __str__(self):
        return (
            f"(NOT ({self.connector}: {', '.join(str(c) for c in self.children)}))"
            if self.negated
            else f"({self.connector}: {', '.join(str(c) for c in self.children)})"
        )

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self}>"

    def __deepcopy__(self, memodict):
        obj = QBase(connector=self.connector, negated=self.negated)
        obj.__class__ = self.__class__
        obj.children = copy.deepcopy(self.children, memodict)
        return obj

    def __len__(self):
        """Return the the number of children this node has."""
        return len(self.children)

    def __bool__(self):
        """Return whether or not this node has children."""
        return bool(self.children)

    def __contains__(self, other):
        """Return True if 'other' is a direct child of this instance."""
        return other in self.children

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        if (self.connector, self.negated) == (other.connector, other.negated):
            return self.children == other.children
        return False

    def __hash__(self):
        return hash(
            (self.__class__, self.connector, self.negated) + tuple(self.children)
        )

    def add(self, data, conn_type, squash=True):
        """
        Combine this tree and the data represented by data using the
        connector conn_type. The combine is done by squashing the node other
        away if possible.

        This tree (self) will never be pushed to a child node of the
        combined tree, nor will the connector or negated properties change.

        Return a node which can be used in place of data regardless if the
        node other got squashed or not.

        If `squash` is False the data is prepared and added as a child to
        this tree without further logic.

        Args:
            conn_type (str, optional ["AND", "OR"]): connection method
        """
        if data in self.children:
            return data
        if not squash:
            self.children.append(data)
            return data
        if self.connector == conn_type:
            # We can reuse self.children to append or squash the node other.
            if (
                isinstance(data, QBase)
                and not data.negated
                and (data.connector == conn_type or len(data) == 1)
            ):
                # We can squash the other node's children directly into this
                # node. We are just doing (AB)(CD) == (ABCD) here, with the
                # addition that if the length of the other node is 1 the
                # connector doesn't matter. However, for the len(self) == 1
                # case we don't want to do the squashing, as it would alter
                # self.connector.
                self.children.extend(data.children)
                return self

            # We could use perhaps additional logic here to see if some
            # children could be used for pushdown here.
            self.children.append(data)
            return data

        obj = self._new_instance(self.children, self.connector, self.negated)
        self.connector = conn_type
        self.children = [obj, data]
        return data

    def negate(self):
        """Negate the sense of the root connector."""
        self.negated = not self.negated


class Q(QBase):
    """
    Encapsulate filters as objects that can then be combined logically (using
    `&` and `|`).
    """

    # Connection types
    AND = "AND"
    OR = "OR"
    default = AND

    def __init__(self, *args, **kwargs):
        connector = kwargs.pop("_connector", None)
        negated = kwargs.pop("_negated", False)
        super().__init__(
            children=list(args) + sorted(kwargs.items()),
            connector=connector,
            negated=negated,
        )

    def _combine(self, other, conn):
        if not isinstance(other, Q):
            raise TypeError(other)

        # If the other Q() is empty, ignore it and just use `self`.
        if not other:
            return copy.deepcopy(self)
        # Or if this Q is empty, ignore it and just use `other`.
        if not self:
            return copy.deepcopy(other)

        obj = type(self)()
        obj.connector = conn
        obj.add(self, conn)
        obj.add(other, conn)
        return obj

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)

    def __invert__(self):
        obj = type(self)()
        obj.add(self, self.AND)
        obj.negate()
        return obj



# pep8: noqa
from neomodel.async_.cardinality import (
    AsyncOne,
    AsyncOneOrMore,
    AsyncZeroOrMore,
    AsyncZeroOrOne,
)
from neomodel.async_.core import AsyncStructuredNode, adb
from neomodel.async_.match import AsyncNodeSet, AsyncTraversal
from neomodel.async_.path import AsyncNeomodelPath
from neomodel.async_.property_manager import AsyncPropertyManager
from neomodel.async_.relationship import AsyncStructuredRel
from neomodel.async_.relationship_manager import (
    AsyncRelationship,
    AsyncRelationshipDefinition,
    AsyncRelationshipFrom,
    AsyncRelationshipManager,
    AsyncRelationshipTo,
)
from neomodel.exceptions import *
from neomodel.match_q import Q  # noqa
from neomodel.properties import (
    AliasProperty,
    ArrayProperty,
    BooleanProperty,
    DateProperty,
    DateTimeFormatProperty,
    DateTimeNeo4jFormatProperty,
    DateTimeProperty,
    EmailProperty,
    FloatProperty,
    FulltextIndex,
    IntegerProperty,
    JSONProperty,
    NormalizedProperty,
    RegexProperty,
    StringProperty,
    UniqueIdProperty,
    VectorIndex,
)
from neomodel.sync_.cardinality import One, OneOrMore, ZeroOrMore, ZeroOrOne
from neomodel.sync_.core import (
    StructuredNode,
    change_neo4j_password,
    clear_neo4j_database,
    db,
    drop_constraints,
    drop_indexes,
    install_all_labels,
    install_labels,
    remove_all_labels,
)
from neomodel.sync_.match import NodeSet, Traversal
from neomodel.sync_.path import NeomodelPath
from neomodel.sync_.property_manager import PropertyManager
from neomodel.sync_.relationship import StructuredRel
from neomodel.sync_.relationship_manager import (
    Relationship,
    RelationshipDefinition,
    RelationshipFrom,
    RelationshipManager,
    RelationshipTo,
)
from neomodel.util import EITHER, INCOMING, OUTGOING

__author__ = "Robin Edwards"
__email__ = "robin.ge@gmail.com"
__license__ = "MIT"
__package__ = "neomodel"



import os

import pytest

from neomodel.util import version_tag_to_integer

NEO4J_URL = os.environ.get("NEO4J_URL", "bolt://localhost:7687")
NEO4J_USERNAME = os.environ.get("NEO4J_USERNAME", "neo4j")
NEO4J_PASSWORD = os.environ.get("NEO4J_PASSWORD", "foobarbaz")


def pytest_addoption(parser):
    """
    Adds the command line option --resetdb.

    :param parser: The parser object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_addoption>`_
    :type Parser object: For more information please see <https://docs.pytest.org/en/latest/reference.html#_pytest.config.Parser>`_
    """
    parser.addoption(
        "--resetdb",
        action="store_true",
        help="Ensures that the database is clear prior to running tests for neomodel",
        default=False,
    )


@pytest.hookimpl
def pytest_collection_modifyitems(items):
    async_items = []
    sync_items = []
    async_connect_to_aura_items = []
    sync_connect_to_aura_items = []
    root_items = []

    for item in items:
        # Check the directory of the item
        directory = item.fspath.dirname.split("/")[-1]
        if directory == "test_contrib":
            directory = item.fspath.dirname.split("/")[-2]

        if "connect_to_aura" in item.name:
            if directory == "async_":
                async_connect_to_aura_items.append(item)
            elif directory == "sync_":
                sync_connect_to_aura_items.append(item)
        else:
            if directory == "async_":
                async_items.append(item)
            elif directory == "sync_":
                sync_items.append(item)
            else:
                root_items.append(item)

    new_order = (
        async_items
        + async_connect_to_aura_items
        + sync_items
        + root_items
        + sync_connect_to_aura_items
    )
    items[:] = new_order


def check_and_skip_neo4j_least_version(required_least_neo4j_version, message):
    """
    Checks if the NEO4J_VERSION is at least `required_least_neo4j_version` and skips a test if not.

    WARNING: If the NEO4J_VERSION variable is not set, this function returns True, allowing the test to go ahead.

    :param required_least_neo4j_version: The least version to check. This must be the numberic representation of the
    version. That is: '3.4.0' would be passed as 340.
    :type required_least_neo4j_version: int
    :param message: An informative message as to why the calling test had to be skipped.
    :type message: str
    :return: A boolean value of True if the version reported is at least `required_least_neo4j_version`
    """
    if (
        "NEO4J_VERSION" in os.environ
        and version_tag_to_integer(os.environ["NEO4J_VERSION"])
        < required_least_neo4j_version
    ):
        pytest.skip(
            "Neo4j version: {}. {}."
            "Skipping test.".format(os.environ["NEO4J_VERSION"], message)
        )



from test._async_compat import mark_sync_test

from pytest import raises, skip

from neomodel import (
    DateProperty,
    IntegerProperty,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    config,
    db,
)
from neomodel.exceptions import (
    NodeClassAlreadyDefined,
    NodeClassNotDefined,
    RelationshipClassRedefined,
)


@mark_sync_test
def test_db_specific_node_labels():
    if not db.edition_is_enterprise():
        skip("Skipping test for community edition")
    db_one = "one"
    db_two = "two"
    db.cypher_query(f"CREATE DATABASE {db_one} IF NOT EXISTS")
    db.cypher_query(f"CREATE DATABASE {db_two} IF NOT EXISTS")

    class Experiment(StructuredNode):
        __label__ = "Experiment"
        name = StringProperty()

    class PatientOne(StructuredNode):
        __label__ = "Patient"
        __target_databases__ = [db_one]
        name = StringProperty()

    class PatientTwo(StructuredNode):
        __label__ = "Patient"
        __target_databases__ = [db_two]
        identifier = StringProperty()

    # This should have reached this point without failing
    # It means db specific registry is allowing reuse of labels in different databases
    # Next test will check if the standard registry still denies reuse of labels
    with raises(NodeClassAlreadyDefined):

        class ExperimentTwo(StructuredNode):
            __label__ = "Experiment"
            name = StringProperty()

        ExperimentTwo(name="experiment2").save()

    # Finally, this tests that db specific registry denies reuse of labels in the same db
    with raises(NodeClassAlreadyDefined):

        class PatientOneBis(StructuredNode):
            __label__ = "Patient"
            __target_databases__ = [db_one]
            name = StringProperty()

        PatientOneBis(name="patient1.2").save()

    # Now, we will test object resolution
    db.close_connection()
    db.set_connection(url=f"{config.DATABASE_URL}/{db_one}")
    db.clear_neo4j_database()
    patient1 = PatientOne(name="patient1").save()
    patients, _ = db.cypher_query("MATCH (n:Patient) RETURN n", resolve_objects=True)
    # This means that the auto object resolution is working
    assert patients[0][0] == patient1

    db.close_connection()
    db.set_connection(url=f"{config.DATABASE_URL}/{db_two}")
    db.clear_neo4j_database()
    patient2 = PatientTwo(identifier="patient2").save()
    patients, _ = db.cypher_query("MATCH (n:Patient) RETURN n", resolve_objects=True)
    assert patients[0][0] == patient2

    db.close_connection()
    db.set_connection(url=config.DATABASE_URL)


@mark_sync_test
def test_resolution_not_defined_class():
    if not db.edition_is_enterprise():
        skip("Skipping test for community edition")

    class PatientX(StructuredNode):
        __label__ = "Patient"
        __target_databases__ = ["db_x"]
        name = StringProperty()

    db.cypher_query("CREATE (n:Gabagool)")
    with raises(
        NodeClassNotDefined,
        match=r"Node with labels Gabagool does not resolve to any of the known objects[\s\S]*Database-specific: db_x.*",
    ):
        _ = db.cypher_query("MATCH (n:Gabagool) RETURN n", resolve_objects=True)



import inspect
import re
import string
import warnings
from dataclasses import dataclass
from typing import Any, AsyncIterator
from typing import Optional as TOptional
from typing import Tuple, Union

from neomodel.async_ import relationship_manager
from neomodel.async_.core import AsyncStructuredNode, adb
from neomodel.async_.relationship import AsyncStructuredRel
from neomodel.semantic_filters import VectorFilter
from neomodel.exceptions import MultipleNodesReturned
from neomodel.match_q import Q, QBase
from neomodel.properties import AliasProperty, ArrayProperty, Property
from neomodel.typing import Subquery, Transformation
from neomodel.util import INCOMING, OUTGOING

CYPHER_ACTIONS_WITH_SIDE_EFFECT_EXPR = re.compile(r"(?i:MERGE|CREATE|DELETE|DETACH)")


def _rel_helper(
    lhs: str,
    rhs: str,
    ident: TOptional[str] = None,
    relation_type: TOptional[str] = None,
    direction: TOptional[int] = None,
    relation_properties: TOptional[dict] = None,
    **kwargs: dict[str, Any],  # NOSONAR
) -> str:
    """
    Generate a relationship matching string, with specified parameters.
    Examples:
    relation_direction = OUTGOING: (lhs)-[relation_ident:relation_type]->(rhs)
    relation_direction = INCOMING: (lhs)<-[relation_ident:relation_type]-(rhs)
    relation_direction = EITHER: (lhs)-[relation_ident:relation_type]-(rhs)

    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param relation_properties: dictionary of relationship properties to match
    :returns: string
    """
    rel_props = ""

    if relation_properties:
        rel_props_str = ", ".join(
            (f"{key}: {value}" for key, value in relation_properties.items())
        )
        rel_props = f" {{{rel_props_str}}}"

    rel_def = ""
    # relation_type is unspecified
    if relation_type is None:
        rel_def = ""
    # all("*" wildcard) relation_type
    elif relation_type == "*":
        rel_def = "[*]"
    else:
        # explicit relation_type
        rel_def = f"[{ident if ident else ''}:`{relation_type}`{rel_props}]"

    stmt = ""
    if direction == OUTGOING:
        stmt = f"-{rel_def}->"
    elif direction == INCOMING:
        stmt = f"<-{rel_def}-"
    else:
        stmt = f"-{rel_def}-"

    # Make sure not to add parenthesis when they are already present
    if lhs[-1] != ")":
        lhs = f"({lhs})"
    if rhs[-1] != ")":
        rhs = f"({rhs})"

    return f"{lhs}{stmt}{rhs}"


def _rel_merge_helper(
    lhs: str,
    rhs: str,
    ident: str = "neomodelident",
    relation_type: TOptional[str] = None,
    direction: TOptional[int] = None,
    relation_properties: TOptional[dict] = None,
    **kwargs: dict[str, Any],  # NOSONAR
) -> str:
    """
    Generate a relationship merging string, with specified parameters.
    Examples:
    relation_direction = OUTGOING: (lhs)-[relation_ident:relation_type]->(rhs)
    relation_direction = INCOMING: (lhs)<-[relation_ident:relation_type]-(rhs)
    relation_direction = EITHER: (lhs)-[relation_ident:relation_type]-(rhs)

    :param lhs: The left hand statement.
    :type lhs: str
    :param rhs: The right hand statement.
    :type rhs: str
    :param ident: A specific identity to name the relationship, or None.
    :type ident: str
    :param relation_type: None for all direct rels, * for all of any length, or a name of an explicit rel.
    :type relation_type: str
    :param direction: None or EITHER for all OUTGOING,INCOMING,EITHER. Otherwise OUTGOING or INCOMING.
    :param relation_properties: dictionary of relationship properties to merge
    :returns: string
    """

    if direction == OUTGOING:
        stmt = "-{0}->"
    elif direction == INCOMING:
        stmt = "<-{0}-"
    else:
        stmt = "-{0}-"

    rel_props = ""
    rel_none_props = ""

    if relation_properties:
        rel_props_str = ", ".join(
            (
                f"{key}: {value}"
                for key, value in relation_properties.items()
                if value is not None
            )
        )
        rel_props = f" {{{rel_props_str}}}"
        if None in relation_properties.values():
            rel_prop_val_str = ", ".join(
                (
                    f"{ident}.{key}=${key!s}"
                    for key, value in relation_properties.items()
                    if value is None
                )
            )
            rel_none_props = (
                f" ON CREATE SET {rel_prop_val_str} ON MATCH SET {rel_prop_val_str}"
            )
    # relation_type is unspecified
    if relation_type is None:
        stmt = stmt.format("")
    # all("*" wildcard) relation_type
    elif relation_type == "*":
        stmt = stmt.format("[*]")
    else:
        # explicit relation_type
        stmt = stmt.format(f"[{ident}:`{relation_type}`{rel_props}]")

    return f"({lhs}){stmt}({rhs}){rel_none_props}"


# special operators
_SPECIAL_OPERATOR_IN = "IN"
_SPECIAL_OPERATOR_ARRAY_IN = "any(x IN {ident}.{prop} WHERE x IN {val})"
_SPECIAL_OPERATOR_INSENSITIVE = "(?i)"
_SPECIAL_OPERATOR_ISNULL = "IS NULL"
_SPECIAL_OPERATOR_ISNOTNULL = "IS NOT NULL"
_SPECIAL_OPERATOR_REGEX = "=~"

_UNARY_OPERATORS = (_SPECIAL_OPERATOR_ISNULL, _SPECIAL_OPERATOR_ISNOTNULL)

_REGEX_INSENSITIVE = _SPECIAL_OPERATOR_INSENSITIVE + "{}"
_REGEX_CONTAINS = ".*{}.*"
_REGEX_STARTSWITH = "{}.*"
_REGEX_ENDSWITH = ".*{}"

# regex operations that require escaping
_STRING_REGEX_OPERATOR_TABLE = {
    "iexact": _REGEX_INSENSITIVE,
    "contains": _REGEX_CONTAINS,
    "icontains": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_CONTAINS,
    "startswith": _REGEX_STARTSWITH,
    "istartswith": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_STARTSWITH,
    "endswith": _REGEX_ENDSWITH,
    "iendswith": _SPECIAL_OPERATOR_INSENSITIVE + _REGEX_ENDSWITH,
}
# regex operations that do not require escaping
_REGEX_OPERATOR_TABLE = {
    "iregex": _REGEX_INSENSITIVE,
}
# list all regex operations, these will require formatting of the value
_REGEX_OPERATOR_TABLE.update(_STRING_REGEX_OPERATOR_TABLE)

# list all supported operators
OPERATOR_TABLE = {
    "lt": "<",
    "gt": ">",
    "lte": "<=",
    "gte": ">=",
    "ne": "<>",
    "in": _SPECIAL_OPERATOR_IN,
    "isnull": _SPECIAL_OPERATOR_ISNULL,
    "regex": _SPECIAL_OPERATOR_REGEX,
    "exact": "=",
}
# add all regex operators
OPERATOR_TABLE.update(_REGEX_OPERATOR_TABLE)

path_split_regex = re.compile(r"__(?!_)|\|")


def install_traversals(
    cls: type[AsyncStructuredNode], node_set: "AsyncNodeSet"
) -> None:
    """
    For a StructuredNode class install Traversal objects for each
    relationship definition on a NodeSet instance
    """
    rels = cls.defined_properties(rels=True, aliases=False, properties=False)

    for key in rels.keys():
        if hasattr(node_set, key):
            raise ValueError(f"Cannot install traversal '{key}' exists on NodeSet")

        rel = getattr(cls, key)
        rel.lookup_node_class()

        traversal = AsyncTraversal(source=node_set, name=key, definition=rel.definition)
        setattr(node_set, key, traversal)


def _handle_special_operators(
    property_obj: Property, key: str, value: str, operator: str, prop: str
) -> Tuple[str, str, str]:
    if operator == _SPECIAL_OPERATOR_IN:
        if not isinstance(value, (list, tuple)):
            raise ValueError(
                f"Value must be a tuple or list for IN operation {key}={value}"
            )
        if isinstance(property_obj, ArrayProperty):
            deflated_value = property_obj.deflate(value)
            operator = _SPECIAL_OPERATOR_ARRAY_IN
        else:
            deflated_value = [property_obj.deflate(v) for v in value]
    elif operator == _SPECIAL_OPERATOR_ISNULL:
        if not isinstance(value, bool):
            raise ValueError(f"Value must be a bool for isnull operation on {key}")
        operator = "IS NULL" if value else "IS NOT NULL"
        deflated_value = None
    elif operator in _REGEX_OPERATOR_TABLE.values():
        deflated_value = property_obj.deflate(value)
        if not isinstance(deflated_value, str):
            raise ValueError(f"Must be a string value for {key}")
        if operator in _STRING_REGEX_OPERATOR_TABLE.values():
            deflated_value = re.escape(deflated_value)
        deflated_value = operator.format(deflated_value)
        operator = _SPECIAL_OPERATOR_REGEX
    else:
        deflated_value = property_obj.deflate(value)

    return deflated_value, operator, prop


def _deflate_value(
    cls: type[AsyncStructuredNode],
    property_obj: Property,
    key: str,
    value: str,
    operator: str,
    prop: str,
) -> Tuple[str, str, str]:
    if isinstance(property_obj, AliasProperty):
        prop = property_obj.aliased_to()
        deflated_value = getattr(cls, prop).deflate(value)
    else:
        # handle special operators
        deflated_value, operator, prop = _handle_special_operators(
            property_obj, key, value, operator, prop
        )

    return deflated_value, operator, prop


def _initialize_filter_args_variables(
    cls: type[AsyncStructuredNode], key: str
) -> Tuple[type[AsyncStructuredNode], None, None, str, bool, str]:
    current_class = cls
    current_rel_model = None
    leaf_prop = None
    operator = "="
    is_rel_property = "|" in key
    prop = key

    return (
        current_class,
        current_rel_model,
        leaf_prop,
        operator,
        is_rel_property,
        prop,
    )


def _process_filter_key(
    cls: type[AsyncStructuredNode], key: str
) -> Tuple[Property, str, str]:
    (
        current_class,
        current_rel_model,
        leaf_prop,
        operator,
        is_rel_property,
        prop,
    ) = _initialize_filter_args_variables(cls, key)

    for part in re.split(path_split_regex, key):
        defined_props = current_class.defined_properties(rels=True)
        # update defined props dictionary with relationship properties if
        # we are filtering by property
        if is_rel_property and current_rel_model:
            defined_props.update(current_rel_model.defined_properties(rels=True))
        if part in defined_props:
            if isinstance(
                defined_props[part], relationship_manager.AsyncRelationshipDefinition
            ):
                defined_props[part].lookup_node_class()
                current_class = defined_props[part].definition["node_class"]
                current_rel_model = defined_props[part].definition["model"]
        elif part in OPERATOR_TABLE:
            operator = OPERATOR_TABLE[part]
            prop, _ = prop.rsplit("__", 1)
            continue
        else:
            raise ValueError(
                f"No such property {part} on {cls.__name__}. Note that Neo4j internals like id or element_id are not allowed for use in this operation."
            )
        leaf_prop = part

    if leaf_prop is None:
        raise ValueError(f"Badly formed filter, no property found in {key}")
    if is_rel_property and current_rel_model:
        property_obj = getattr(current_rel_model, leaf_prop)
    else:
        property_obj = getattr(current_class, leaf_prop)

    return property_obj, operator, prop


def process_filter_args(cls: type[AsyncStructuredNode], kwargs: dict[str, Any]) -> dict:
    """
    loop through properties in filter parameters check they match class definition
    deflate them and convert into something easy to generate cypher from
    """
    output = {}

    for key, value in kwargs.items():
        property_obj, operator, prop = _process_filter_key(cls, key)
        deflated_value, operator, prop = _deflate_value(
            cls, property_obj, key, value, operator, prop
        )
        # map property to correct property name in the database
        db_property = prop

        output[db_property] = (operator, deflated_value)
    return output


def process_has_args(
    cls: type[AsyncStructuredNode], kwargs: dict[str, Any]
) -> tuple[dict, dict]:
    """
    loop through has parameters check they correspond to class rels defined
    """
    rel_definitions = cls.defined_properties(properties=False, rels=True, aliases=False)

    match, dont_match = {}, {}

    for key, value in kwargs.items():
        if key not in rel_definitions:
            raise ValueError(f"No such relation {key} defined on a {cls.__name__}")

        rhs_ident = key

        rel_definitions[key].lookup_node_class()

        if value is True:
            match[rhs_ident] = rel_definitions[key].definition
        elif value is False:
            dont_match[rhs_ident] = rel_definitions[key].definition
        elif isinstance(value, AsyncNodeSet):
            raise NotImplementedError("Not implemented yet")
        else:
            raise ValueError("Expecting True / False / NodeSet got: " + repr(value))

    return match, dont_match


class QueryAST:
    match: list[str]
    optional_match: list[str]
    where: list[str]
    with_clause: TOptional[str]
    return_clause: TOptional[str]
    order_by: TOptional[list[str]]
    skip: TOptional[int]
    limit: TOptional[int]
    result_class: TOptional[type]
    lookup: TOptional[str]
    additional_return: TOptional[list[str]]
    is_count: TOptional[bool]
    vector_index_query: TOptional[type]
    def __init__(
        self,
        match: TOptional[list[str]] = None,
        optional_match: TOptional[list[str]] = None,
        where: TOptional[list[str]] = None,
        optional_where: TOptional[list[str]] = None,
        with_clause: TOptional[str] = None,
        return_clause: TOptional[str] = None,
        order_by: TOptional[list[str]] = None,
        skip: TOptional[int] = None,
        limit: TOptional[int] = None,
        result_class: TOptional[type] = None,
        lookup: TOptional[str] = None,
        additional_return: TOptional[list[str]] = None,
        is_count: TOptional[bool] = False,
        vector_index_query: TOptional[type] = None,
    ) -> None:
        self.match = match if match else []
        self.optional_match = optional_match if optional_match else []
        self.where = where if where else []
        self.optional_where = optional_where if optional_where else []
        self.with_clause = with_clause
        self.return_clause = return_clause
        self.order_by = order_by
        self.skip = skip
        self.limit = limit
        self.result_class = result_class
        self.lookup = lookup
        self.additional_return: list[str] = (
            additional_return if additional_return else []
        )
        self.is_count = is_count
        self.vector_index_query = vector_index_query
        self.subgraph: dict = {}
        self.mixed_filters: bool = False


class AsyncQueryBuilder:
    def __init__(
        self, node_set: "AsyncBaseSet", subquery_namespace: TOptional[str] = None
    ) -> None:
        self.node_set = node_set
        self._ast = QueryAST()
        self._query_params: dict = {}
        self._place_holder_registry: dict = {}
        self._relation_identifier_count: int = 0
        self._node_identifier_count: int = 0
        self._subquery_namespace: TOptional[str] = subquery_namespace

    async def build_ast(self) -> "AsyncQueryBuilder":
        if isinstance(self.node_set, AsyncNodeSet) and hasattr(
            self.node_set, "relations_to_fetch"
        ):
            for relation in self.node_set.relations_to_fetch:
                self.build_traversal_from_path(relation, self.node_set.source)
        
        if isinstance(self.node_set, AsyncNodeSet) and hasattr(self.node_set, "_vector_query"):
            if self.node_set._vector_query:
                self.build_vector_query(self.node_set._vector_query, self.node_set.source)

        await self.build_source(self.node_set)

        if hasattr(self.node_set, "skip"):
            self._ast.skip = self.node_set.skip
        if hasattr(self.node_set, "limit"):
            self._ast.limit = self.node_set.limit

        return self

    async def build_source(
        self, source: Union["AsyncTraversal", "AsyncNodeSet", AsyncStructuredNode, Any]
    ) -> str:
        if isinstance(source, AsyncTraversal):
            return await self.build_traversal(source)
        if isinstance(source, AsyncNodeSet):
            if inspect.isclass(source.source) and issubclass(
                source.source, AsyncStructuredNode
            ):
                ident = self.build_label(source.source.__label__.lower(), source.source)
            else:
                ident = await self.build_source(source.source)

            self.build_additional_match(ident, source)

            if hasattr(source, "order_by_elements"):
                self.build_order_by(ident, source)

            # source.filters seems to be used only by Traversal objects
            # source.q_filters is used by NodeSet objects
            if source.filters or source.q_filters:
                self.build_where_stmt(
                    ident=ident,
                    filters=source.filters,
                    source_class=source.source_class,
                    q_filters=source.q_filters,
                )

            return ident
        if isinstance(source, AsyncStructuredNode):
            return await self.build_node(source)
        raise ValueError("Unknown source type " + repr(source))

    def create_relation_identifier(self) -> str:
        self._relation_identifier_count += 1
        return f"r{self._relation_identifier_count}"

    def create_node_identifier(self, prefix: str, path: str) -> str:
        if path not in self.node_set._unique_variables:
            self._node_identifier_count += 1
            return f"{prefix}{self._node_identifier_count}"
        return prefix

    def build_order_by(self, ident: str, source: "AsyncNodeSet") -> None:
        if "?" in source.order_by_elements:
            self._ast.with_clause = f"{ident}, rand() as r"
            self._ast.order_by = ["r"]
        else:
            order_by = []
            for elm in source.order_by_elements:
                if isinstance(elm, RawCypher):
                    order_by.append(elm.render({"n": ident}))
                    continue
                is_rel_property = "|" in elm
                if "__" not in elm and not is_rel_property:
                    prop = elm.split(" ")[0] if " " in elm else elm
                    if prop not in source.source_class.defined_properties(rels=False):
                        raise ValueError(
                            f"No such property {prop} on {source.source_class.__name__}. "
                            f"Note that Neo4j internals like id or element_id are not allowed "
                            f"for use in this operation."
                        )
                    order_by.append(f"{ident}.{elm}")
                else:
                    path, prop = elm.rsplit("__" if not is_rel_property else "|", 1)
                    result = self.lookup_query_variable(
                        path, return_relation=is_rel_property
                    )
                    if result:
                        order_by.append(f"{result[0]}.{prop}")
            self._ast.order_by = order_by


    def build_vector_query(self, vectorfilter: "VectorFilter", source: "NodeSet"):
        """
        Query a vector indexed property on the node. 
        """
        try:
            attribute = getattr(source, vectorfilter.vector_attribute_name)
        except AttributeError:
            raise # This raises the base AttributeError and provides potential correction

        if not attribute.vector_index:
            raise AttributeError(f"Attribute {vectorfilter.vector_attribute_name} is not declared with a vector index.")

        vectorfilter.index_name = f"vector_index_{source.__label__}_{vectorfilter.vector_attribute_name}"
        vectorfilter.nodeSetLabel = source.__label__.lower()

        self._ast.vector_index_query = vectorfilter
        self._ast.return_clause = f"{vectorfilter.nodeSetLabel}, score"
        self._ast.result_class = source.__class__


    async def build_traversal(self, traversal: "AsyncTraversal") -> str:
        """
        traverse a relationship from a node to a set of nodes
        """
        # build source
        rhs_label = ":" + traversal.target_class.__label__

        # build source
        rel_ident = self.create_relation_identifier()
        lhs_ident = await self.build_source(traversal.source)
        traversal_ident = f"{traversal.name}_{rel_ident}"
        rhs_ident = traversal_ident + rhs_label
        self._ast.return_clause = traversal_ident
        self._ast.result_class = traversal.target_class

        stmt = _rel_helper(
            lhs=lhs_ident,
            rhs=rhs_ident,
            ident=rel_ident,
            **traversal.definition,
        )
        self._ast.match.append(stmt)

        if traversal.filters:
            self.build_where_stmt(rel_ident, traversal.filters, traversal.source_class)

        return traversal_ident

    def _additional_return(self, name: str) -> None:
        if (
            not self._ast.additional_return or name not in self._ast.additional_return
        ) and name != self._ast.return_clause:
            if not self._ast.additional_return:
                self._ast.additional_return = []
            self._ast.additional_return.append(name)

    def build_traversal_from_path(
        self, relation: "Path", source_class: Any
    ) -> Tuple[str, Any]:
        path: str = relation.value
        stmt: str = ""
        source_class_iterator = source_class
        parts = re.split(path_split_regex, path)
        subgraph = self._ast.subgraph
        rel_iterator: str = ""
        already_present = False
        existing_rhs_name = ""
        for index, part in enumerate(parts):
            relationship = getattr(source_class_iterator, part)
            if rel_iterator:
                rel_iterator += "__"
            rel_iterator += part
            # build source
            if "node_class" not in relationship.definition:
                relationship.lookup_node_class()
            if not stmt:
                lhs_label = source_class_iterator.__label__
                lhs_name = lhs_label.lower()
                lhs_ident = f"{lhs_name}:{lhs_label}"
                if not index:
                    # This is the first one, we make sure that 'return'
                    # contains the primary node so _contains() works
                    # as usual
                    self._ast.return_clause = lhs_name
                    if self._subquery_namespace:
                        # Don't include label in identifier if we are in a subquery
                        lhs_ident = lhs_name
                elif relation.include_nodes_in_return:
                    self._additional_return(lhs_name)
            else:
                lhs_ident = stmt

            already_present = part in subgraph
            rel_ident = self.create_relation_identifier()
            rhs_label = relationship.definition["node_class"].__label__
            if relation.relation_filtering:
                rhs_name = rel_ident
                rhs_ident = f":{rhs_label}"
            else:
                if index + 1 == len(parts) and relation.alias:
                    # If an alias is defined, use it to store the last hop in the path
                    rhs_name = relation.alias
                else:
                    rhs_name = f"{rhs_label.lower()}_{rel_iterator}"
                    rhs_name = self.create_node_identifier(rhs_name, rel_iterator)
                rhs_ident = f"{rhs_name}:{rhs_label}"

            if relation.include_nodes_in_return and not already_present:
                self._additional_return(rhs_name)

            if not already_present:
                subgraph[part] = {
                    "target": relationship.definition["node_class"],
                    "children": {},
                    "variable_name": rhs_name,
                    "rel_variable_name": rel_ident,
                }
            else:
                existing_rhs_name = subgraph[part][
                    (
                        "rel_variable_name"
                        if relation.relation_filtering
                        else "variable_name"
                    )
                ]
            if relation.include_rels_in_return and not already_present:
                self._additional_return(rel_ident)
            stmt = _rel_helper(
                lhs=lhs_ident,
                rhs=rhs_ident,
                ident=rel_ident,
                direction=relationship.definition["direction"],
                relation_type=relationship.definition["relation_type"],
            )
            source_class_iterator = relationship.definition["node_class"]
            subgraph = subgraph[part]["children"]

        if not already_present:
            if relation.optional:
                self._ast.optional_match.append(stmt)
            else:
                self._ast.match.append(stmt)
            return rhs_name, relationship.definition["node_class"]

        return existing_rhs_name, relationship.definition["node_class"]

    async def build_node(self, node: AsyncStructuredNode) -> str:
        ident = node.__class__.__name__.lower()
        place_holder = self._register_place_holder(ident)

        # Hack to emulate START to lookup a node by id
        _node_lookup = f"MATCH ({ident}) WHERE {await adb.get_id_method()}({ident})=${place_holder} WITH {ident}"
        self._ast.lookup = _node_lookup

        self._query_params[place_holder] = await adb.parse_element_id(node.element_id)

        self._ast.return_clause = ident
        self._ast.result_class = node.__class__
        return ident

    def build_label(self, ident: str, cls: type[AsyncStructuredNode]) -> str:
        """
        match nodes by a label
        """
        ident_w_label = ident + ":" + cls.__label__

        if not self._ast.return_clause:
            if (
                not self._ast.additional_return
                or ident not in self._ast.additional_return
            ):
                self._ast.match.append(f"({ident_w_label})")
                self._ast.return_clause = ident
                self._ast.result_class = cls
        elif not self._ast.match:
            # If we get here, it means return_clause was filled because of an
            # optional match, so we add a regular match for root node.
            # Not very elegant, this part would deserve a refactoring...
            self._ast.match.append(f"({ident_w_label})")
            self._ast.result_class = cls
        return ident

    def build_additional_match(self, ident: str, node_set: "AsyncNodeSet") -> None:
        """
        handle additional matches supplied by 'has()' calls
        """
        source_ident = ident

        for _, value in node_set.must_match.items():
            if isinstance(value, dict):
                label = ":" + value["node_class"].__label__
                stmt = f"EXISTS ({_rel_helper(lhs=source_ident, rhs=label, ident='', **value)})"
                self._ast.where.append(stmt)
            else:
                raise ValueError("Expecting dict got: " + repr(value))

        for _, val in node_set.dont_match.items():
            if isinstance(val, dict):
                label = ":" + val["node_class"].__label__
                stmt = f"NOT EXISTS ({_rel_helper(lhs=source_ident, rhs=label, ident='', **val)})"
                self._ast.where.append(stmt)
            else:
                raise ValueError("Expecting dict got: " + repr(val))

    def _register_place_holder(self, key: str) -> str:
        if key in self._place_holder_registry:
            self._place_holder_registry[key] += 1
        else:
            self._place_holder_registry[key] = 1
        place_holder = f"{key}_{self._place_holder_registry[key]}"
        if self._subquery_namespace:
            place_holder = f"{self._subquery_namespace}_{place_holder}"
        return place_holder

    def _parse_path(
        self, source_class: type[AsyncStructuredNode], prop: str
    ) -> Tuple[str, str, str, Any, bool]:
        is_rel_filter = "|" in prop
        if is_rel_filter:
            path, prop = prop.rsplit("|", 1)
        else:
            path, prop = prop.rsplit("__", 1)
        result = self.lookup_query_variable(path, return_relation=is_rel_filter)
        is_optional_relation = False
        if not result:
            ident, target_class = self.build_traversal_from_path(
                Path(
                    value=path,
                    relation_filtering=is_rel_filter,
                ),
                source_class,
            )
        else:
            ident, target_class, is_optional_relation = result
        return ident, path, prop, target_class, is_optional_relation

    def _finalize_filter_statement(
        self, operator: str, ident: str, prop: str, val: Any
    ) -> str:
        if operator in _UNARY_OPERATORS:
            # unary operators do not have a parameter
            statement = f"{ident}.{prop} {operator}"
        else:
            place_holder = self._register_place_holder(ident + "_" + prop)
            if operator == _SPECIAL_OPERATOR_ARRAY_IN:
                statement = operator.format(
                    ident=ident,
                    prop=prop,
                    val=f"${place_holder}",
                )
            else:
                statement = f"{ident}.{prop} {operator} ${place_holder}"
            self._query_params[place_holder] = val

        return statement

    def _build_filter_statements(
        self,
        ident: str,
        filters: dict[str, tuple],
        target: list[tuple[str, bool]],
        source_class: type[AsyncStructuredNode],
    ) -> None:
        for prop, op_and_val in filters.items():
            path = None
            is_rel_filter = "|" in prop
            target_class = source_class
            is_optional_relation = False
            if "__" in prop or is_rel_filter:
                (
                    ident,
                    path,
                    prop,
                    target_class,
                    is_optional_relation,
                ) = self._parse_path(source_class, prop)
            operator, val = op_and_val
            if not is_rel_filter:
                prop = target_class.defined_properties(rels=False)[
                    prop
                ].get_db_property_name(prop)
            statement = self._finalize_filter_statement(operator, ident, prop, val)
            target.append((statement, is_optional_relation))

    def _parse_q_filters(
        self, ident: str, q: Union[QBase, Any], source_class: type[AsyncStructuredNode]
    ) -> tuple[str, str]:
        target: list[tuple[str, bool]] = []

        def add_to_target(statement: str, connector: Q, optional: bool) -> None:
            if not statement:
                return
            if connector == Q.OR:
                statement = f"({statement})"
            target.append((statement, optional))

        for child in q.children:
            if isinstance(child, QBase):
                q_childs, q_opt_childs = self._parse_q_filters(
                    ident, child, source_class
                )
                add_to_target(q_childs, child.connector, False)
                add_to_target(q_opt_childs, child.connector, True)
            else:
                kwargs = {child[0]: child[1]}
                filters = process_filter_args(source_class, kwargs)
                self._build_filter_statements(ident, filters, target, source_class)
        match_filters = [filter[0] for filter in target if not filter[1]]
        opt_match_filters = [filter[0] for filter in target if filter[1]]
        if q.connector == Q.OR and match_filters and opt_match_filters:
            # In this case, we can't split filters in two WHERE statements so we move
            # everything into the one applied after OPTIONAL MATCH statements...
            opt_match_filters += match_filters
            match_filters = []
            self._ast.mixed_filters = True

        ret = f" {q.connector} ".join(match_filters)
        if ret and q.negated:
            ret = f"NOT ({ret})"
        opt_ret = f" {q.connector} ".join(opt_match_filters)
        if opt_ret and q.negated:
            opt_ret = f"NOT ({opt_ret})"
        return ret, opt_ret

    def build_where_stmt(
        self,
        ident: str,
        filters: list,
        source_class: type[AsyncStructuredNode],
        q_filters: Union[QBase, Any, None] = None,
    ) -> None:
        """
        Construct a where statement from some filters.

        We make a difference between filters applied to variables coming from MATCH and
        OPTIONAL MATCH statements.

        """
        if q_filters is not None:
            stmt, opt_stmt = self._parse_q_filters(ident, q_filters, source_class)
            if stmt:
                self._ast.where.append(stmt)
            if opt_stmt:
                self._ast.optional_where.append(opt_stmt)
        else:
            stmts = []
            for row in filters:
                negate = False

                # pre-process NOT cases as they are nested dicts
                if "__NOT__" in row and len(row) == 1:
                    negate = True
                    row = row["__NOT__"]

                for prop, operator_and_val in row.items():
                    operator, val = operator_and_val
                    if operator in _UNARY_OPERATORS:
                        # unary operators do not have a parameter
                        statement = (
                            f"{'NOT' if negate else ''} {ident}.{prop} {operator}"
                        )
                    else:
                        place_holder = self._register_place_holder(ident + "_" + prop)
                        statement = f"{'NOT' if negate else ''} {ident}.{prop} {operator} ${place_holder}"
                        self._query_params[place_holder] = val
                    stmts.append(statement)

            self._ast.where.append(" AND ".join(stmts))

    def lookup_query_variable(
        self, path: str, return_relation: bool = False
    ) -> TOptional[Tuple[str, Any, bool]]:
        """Retrieve the variable name generated internally for the given traversal path."""
        subgraph = self._ast.subgraph
        if not subgraph:
            return None
        traversals = re.split(path_split_regex, path)
        if len(traversals) == 0:
            raise ValueError("Can only lookup traversal variables")
        if traversals[0] not in subgraph:
            return None

        # Check if relation is coming from an optional MATCH
        # (declared using fetch|traverse_relations)
        is_optional_relation = False
        for relation in self.node_set.relations_to_fetch:
            if relation.value == path:
                is_optional_relation = relation.optional
                break

        subgraph = subgraph[traversals[0]]
        if len(traversals) == 1:
            variable_to_return = f"{subgraph['rel_variable_name' if return_relation else 'variable_name']}"
            return variable_to_return, subgraph["target"], is_optional_relation
        variable_to_return = ""
        last_property = traversals[-1]
        for part in traversals[1:]:
            child = subgraph["children"].get(part)
            if not child:
                return None
            subgraph = child
            if part == last_property:
                # if last part of prop is the last traversal
                # we are safe to lookup the variable from the query
                variable_to_return = f"{subgraph['rel_variable_name' if return_relation else 'variable_name']}"
        return variable_to_return, subgraph["target"], is_optional_relation

    def build_query(self) -> str:
        query: str = ""

        if self._ast.lookup:
            query += self._ast.lookup

        if self._ast.vector_index_query:

            query += f"""CALL () {{ 
                CALL db.index.vector.queryNodes("{self._ast.vector_index_query.index_name}", {self._ast.vector_index_query.topk}, {self._ast.vector_index_query.vector}) 
                YIELD node AS {self._ast.vector_index_query.nodeSetLabel}, score 
                RETURN {self._ast.vector_index_query.nodeSetLabel}, score 
                }}"""

            # This ensures that we bring the context of the new nodeSet and score along with us for metadata filtering
            query += f""" WITH {self._ast.vector_index_query.nodeSetLabel}, score"""

        # Instead of using only one MATCH statement for every relation
        # to follow, we use one MATCH per relation (to avoid cartesian
        # product issues...).
        # There might be optimizations to be done, using projections,
        # or pusing patterns instead of a chain of OPTIONAL MATCH.
        if self._ast.match:
            query += " MATCH "
            query += " MATCH ".join(i for i in self._ast.match)

        if self._ast.where:
            query += " WHERE "
            query += " AND ".join(self._ast.where)

        if self._ast.optional_match:
            query += " OPTIONAL MATCH "
            query += " OPTIONAL MATCH ".join(i for i in self._ast.optional_match)

        if self._ast.optional_where:
            if self._ast.mixed_filters:
                # Make sure filtering works as expected with optional match, even if it's not performant...
                query += " WITH *"
            query += " WHERE "
            query += " AND ".join(self._ast.optional_where)

        if self._ast.with_clause:
            query += " WITH "
            query += self._ast.with_clause

        returned_items: list[str] = []
        if hasattr(self.node_set, "_subqueries"):
            for subquery in self.node_set._subqueries:
                query += " CALL {"
                if subquery["initial_context"]:
                    query += " WITH "
                    context: list[str] = []
                    for var in subquery["initial_context"]:
                        if isinstance(var, (NodeNameResolver, RelationNameResolver)):
                            context.append(var.resolve(self))
                        else:
                            context.append(var)
                    query += ",".join(context)

                query += f"{subquery['query']} }} "
                self._query_params.update(subquery["query_params"])
                for varname in subquery["return_set"]:
                    # We declare the returned variables as "virtual" relations of the
                    # root node class to make sure they will be translated by a call to
                    # resolve_subgraph() (otherwise, they will be lost).
                    # This is probably a temporary solution until we find something better...
                    self._ast.subgraph[varname] = {
                        "target": None,  # We don't need target class in this use case
                        "children": {},
                        "variable_name": varname,
                        "rel_variable_name": varname,
                    }
                returned_items += subquery["return_set"]

        if hasattr(self.node_set, "_intermediate_transforms"):
            for transform in self.node_set._intermediate_transforms:
                query += " WITH "
                query += "DISTINCT " if transform.get("distinct") else ""
                injected_vars: list = []
                # Reset return list since we'll probably invalidate most variables
                self._ast.return_clause = ""
                self._ast.additional_return = []
                returned_items = []
                for name, varprops in transform["vars"].items():
                    source = varprops["source"]
                    if isinstance(source, (NodeNameResolver, RelationNameResolver)):
                        transformation = source.resolve(self)
                    else:
                        transformation = source
                    if varprops.get("source_prop"):
                        transformation += f".{varprops['source_prop']}"
                    transformation += f" AS {name}"
                    if varprops.get("include_in_return"):
                        returned_items += [name]
                    injected_vars.append(transformation)
                query += ",".join(injected_vars)
                if not transform["ordering"]:
                    continue
                query += " ORDER BY "
                ordering: list = []
                for item in transform["ordering"]:
                    if isinstance(item, RawCypher):
                        ordering.append(item.render({}))
                        continue
                    if item.startswith("-"):
                        ordering.append(f"{item[1:]} DESC")
                    else:
                        ordering.append(item)
                query += ",".join(ordering)

        query += " RETURN "
        if self._ast.return_clause and not self._subquery_namespace:
            returned_items.append(self._ast.return_clause)
        if self._ast.additional_return:
            returned_items += self._ast.additional_return
        if hasattr(self.node_set, "_extra_results"):
            for props in self.node_set._extra_results:
                leftpart = props["vardef"].render(self)
                varname = (
                    props["alias"]
                    if props.get("alias")
                    else props["vardef"].get_internal_name()
                )
                if varname in returned_items:
                    # We're about to override an existing variable, delete it first to
                    # avoid duplicate error
                    returned_items.remove(varname)
                returned_items.append(f"{leftpart} AS {varname}")

        query += ", ".join(returned_items)

        if self._ast.order_by:
            query += " ORDER BY "
            query += ", ".join(self._ast.order_by)

        # If we return a count with pagination, pagination has to happen before RETURN
        # It will then be included in the WITH clause already
        if self._ast.skip and not self._ast.is_count:
            query += f" SKIP {self._ast.skip}"

        if self._ast.limit and not self._ast.is_count:
            query += f" LIMIT {self._ast.limit}"

        return query

    async def _count(self) -> int:
        self._ast.is_count = True
        # If we return a count with pagination, pagination has to happen before RETURN
        # Like : WITH my_var SKIP 10 LIMIT 10 RETURN count(my_var)
        self._ast.with_clause = f"{self._ast.return_clause}"
        if self._ast.skip:
            self._ast.with_clause += f" SKIP {self._ast.skip}"

        if self._ast.limit:
            self._ast.with_clause += f" LIMIT {self._ast.limit}"

        self._ast.return_clause = f"count({self._ast.return_clause})"
        # drop order_by, results in an invalid query
        self._ast.order_by = None
        # drop additional_return to avoid unexpected result
        self._ast.additional_return = None
        query = self.build_query()
        results, _ = await adb.cypher_query(query, self._query_params)
        return int(results[0][0])

    async def _contains(self, node_element_id: TOptional[Union[str, int]]) -> bool:
        # inject id = into ast
        if not self._ast.return_clause and self._ast.additional_return:
            self._ast.return_clause = self._ast.additional_return[0]
        if not self._ast.return_clause:
            raise ValueError("Cannot use contains without a return clause")
        ident = self._ast.return_clause
        place_holder = self._register_place_holder(ident + "_contains")
        self._ast.where.append(
            f"{await adb.get_id_method()}({ident}) = ${place_holder}"
        )
        self._query_params[place_holder] = node_element_id
        return await self._count() >= 1

    async def _execute(self, lazy: bool = False, dict_output: bool = False) -> Any:
        if lazy:
            # inject id() into return or return_set
            if self._ast.return_clause:
                self._ast.return_clause = (
                    f"{await adb.get_id_method()}({self._ast.return_clause})"
                )
            else:
                if self._ast.additional_return is not None:
                    self._ast.additional_return = [
                        f"{await adb.get_id_method()}({item})"
                        for item in self._ast.additional_return
                    ]
        query = self.build_query()
        results, prop_names = await adb.cypher_query(
            query,
            self._query_params,
            resolve_objects=True,
        )
        if dict_output:
            for item in results:
                yield dict(zip(prop_names, item))
            return
        # The following is not as elegant as it could be but had to be copied from the
        # version prior to cypher_query with the resolve_objects capability.
        # It seems that certain calls are only supposed to be focusing to the first
        # result item returned (?)
        if results and len(results[0]) == 1:
            for n in results:
                yield n[0]
        else:
            for result in results:
                yield result


class AsyncBaseSet:
    """
    Base class for all node sets.

    Contains common python magic methods, __len__, __contains__ etc
    """

    query_cls = AsyncQueryBuilder
    source_class: type[AsyncStructuredNode]

    async def all(self, lazy: bool = False) -> list:
        """
        Return all nodes belonging to the set
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :return: list of nodes
        :rtype: list
        """
        ast = await self.query_cls(self).build_ast()
        results = [
            node async for node in ast._execute(lazy)
        ]  # Collect all nodes asynchronously
        return results

    async def __aiter__(self) -> AsyncIterator:
        ast = await self.query_cls(self).build_ast()
        async for item in ast._execute():
            yield item

    async def get_len(self) -> int:
        ast = await self.query_cls(self).build_ast()
        return await ast._count()

    async def check_bool(self) -> bool:
        """
        Override for __bool__ dunder method.
        :return: True if the set contains any nodes, False otherwise
        :rtype: bool
        """
        ast = await self.query_cls(self).build_ast()
        _count = await ast._count()
        return _count > 0

    async def check_nonzero(self) -> bool:
        """
        Override for __bool__ dunder method.
        :return: True if the set contains any node, False otherwise
        :rtype: bool
        """
        return await self.check_bool()

    async def check_contains(self, obj: Union[AsyncStructuredNode, Any]) -> bool:
        if isinstance(obj, AsyncStructuredNode):
            if hasattr(obj, "element_id") and obj.element_id is not None:
                ast = await self.query_cls(self).build_ast()
                obj_element_id = await adb.parse_element_id(obj.element_id)
                return await ast._contains(obj_element_id)
            raise ValueError("Unsaved node: " + repr(obj))

        raise ValueError("Expecting StructuredNode instance")

    async def get_item(self, key: Union[int, slice]) -> TOptional["AsyncBaseSet"]:
        if isinstance(key, slice):
            if key.stop and key.start:
                self.limit = key.stop - key.start
                self.skip = key.start
            elif key.stop:
                self.limit = key.stop
            elif key.start:
                self.skip = key.start

            return self

        if isinstance(key, int):
            self.skip = key
            self.limit = 1

            ast = await self.query_cls(self).build_ast()
            _first_item = [node async for node in ast._execute()][0]
            return _first_item

        return None


@dataclass
class Optional:  # type: ignore[no-redef]
    """Simple relation qualifier."""

    relation: str


@dataclass
class Path:
    """Path traversal definition."""

    value: str
    optional: bool = False
    include_nodes_in_return: bool = True
    include_rels_in_return: bool = True
    relation_filtering: bool = False
    alias: TOptional[str] = None


@dataclass
class RelationNameResolver:
    """Helper to refer to a relation variable name.

    Since variable names are generated automatically within MATCH statements (for
    anything injected using fetch_relations or traverse_relations), we need a way to
    retrieve them.

    """

    relation: str

    def resolve(self, qbuilder: AsyncQueryBuilder) -> str:
        result = qbuilder.lookup_query_variable(self.relation, True)
        if result is None:
            raise ValueError(
                f"Unable to resolve variable name for relation {self.relation}"
            )
        return result[0]


@dataclass
class NodeNameResolver:
    """Helper to refer to a node variable name.

    Since variable names are generated automatically within MATCH statements (for
    anything injected using fetch_relations or traverse_relations), we need a way to
    retrieve them.

    """

    node: str

    def resolve(self, qbuilder: AsyncQueryBuilder) -> str:
        if self.node == "self" and qbuilder._ast.return_clause:
            return qbuilder._ast.return_clause
        result = qbuilder.lookup_query_variable(self.node)
        if result is None:
            raise ValueError(f"Unable to resolve variable name for node {self.node}")
        return result[0]


@dataclass
class BaseFunction:
    input_name: Union[str, "BaseFunction", NodeNameResolver, RelationNameResolver]

    def __post_init__(self) -> None:
        self._internal_name: str = ""

    def get_internal_name(self) -> str:
        return self._internal_name

    def resolve_internal_name(self, qbuilder: AsyncQueryBuilder) -> str:
        if isinstance(self.input_name, (NodeNameResolver, RelationNameResolver)):
            self._internal_name = self.input_name.resolve(qbuilder)
        else:
            self._internal_name = str(self.input_name)
        return self._internal_name

    def render(self, qbuilder: AsyncQueryBuilder) -> str:
        raise NotImplementedError


@dataclass
class AggregatingFunction(BaseFunction):
    """Base aggregating function class."""

    pass


@dataclass
class Collect(AggregatingFunction):
    """collect() function."""

    distinct: bool = False

    def render(self, qbuilder: AsyncQueryBuilder) -> str:
        varname = self.resolve_internal_name(qbuilder)
        if self.distinct:
            return f"collect(DISTINCT {varname})"
        return f"collect({varname})"


@dataclass
class ScalarFunction(BaseFunction):
    """Base scalar function class."""

    @property
    def function_name(self) -> str:
        raise NotImplementedError

    def render(self, qbuilder: AsyncQueryBuilder) -> str:
        if isinstance(self.input_name, str):
            content = str(self.input_name)
        elif isinstance(self.input_name, BaseFunction):
            content = self.input_name.render(qbuilder)
            self._internal_name = self.input_name.get_internal_name()
        else:
            content = self.resolve_internal_name(qbuilder)
        return f"{self.function_name}({content})"


@dataclass
class Last(ScalarFunction):
    """last() function."""

    @property
    def function_name(self) -> str:
        return "last"


@dataclass
class Size(ScalarFunction):
    """size() function."""

    @property
    def function_name(self) -> str:
        return "size"


@dataclass
class RawCypher:
    """Helper to inject raw cypher statement.

    Can be used in order_by() call for example.

    """

    statement: str

    def __post_init__(self) -> None:
        if CYPHER_ACTIONS_WITH_SIDE_EFFECT_EXPR.search(self.statement):
            raise ValueError(
                "RawCypher: Do not include any action that has side effect"
            )

    def render(self, context: dict) -> str:
        return string.Template(self.statement).substitute(context)


class AsyncNodeSet(AsyncBaseSet):
    """
    A class representing as set of nodes matching common query parameters
    """

    def __init__(self, source: Any) -> None:
        self.source = source  # could be a Traverse object or a node class
        if isinstance(source, AsyncTraversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, AsyncStructuredNode):
            self.source_class = source
        elif isinstance(source, AsyncStructuredNode):
            self.source_class = source.__class__
        else:
            raise ValueError("Bad source for nodeset " + repr(source))

        # setup Traversal objects using relationship definitions
        install_traversals(self.source_class, self)

        self.filters: list = []
        self.q_filters = Q()
        self.order_by_elements: list = []

        # used by has()
        self.must_match: dict = {}
        self.dont_match: dict = {}

        self.relations_to_fetch: list[Path] = []
        self._extra_results: list = []
        self._subqueries: list[Subquery] = []
        self._intermediate_transforms: list = []
        self._unique_variables: list[str] = []
        self._vector_query: str = None

    def __await__(self) -> Any:
        return self.all().__await__()  # type: ignore[attr-defined]

    async def _get(
        self, limit: TOptional[int] = None, lazy: bool = False, **kwargs: dict[str, Any]
    ) -> list:
        self.filter(**kwargs)
        if limit:
            self.limit = limit
        ast = await self.query_cls(self).build_ast()
        results = [node async for node in ast._execute(lazy)]
        return results

    async def get(self, lazy: bool = False, **kwargs: Any) -> Any:
        """
        Retrieve one node from the set matching supplied parameters
        :param lazy: False by default, specify True to get nodes with id only without the parameters.
        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = await self._get(limit=2, lazy=lazy, **kwargs)
        if len(result) > 1:
            raise MultipleNodesReturned(repr(kwargs))
        if not result:
            raise self.source_class.DoesNotExist(repr(kwargs))
        return result[0]

    async def get_or_none(self, **kwargs: Any) -> Any:
        """
        Retrieve a node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return await self.get(**kwargs)
        except self.source_class.DoesNotExist:
            return None

    async def first(self, **kwargs: Any) -> Any:
        """
        Retrieve the first node from the set matching supplied parameters

        :param kwargs: same syntax as `filter()`
        :return: node
        """
        result = await self._get(limit=1, **kwargs)
        if result:
            return result[0]
        else:
            raise self.source_class.DoesNotExist(repr(kwargs))

    async def first_or_none(self, **kwargs: Any) -> Any:
        """
        Retrieve the first node from the set matching supplied parameters or return none

        :param kwargs: same syntax as `filter()`
        :return: node or none
        """
        try:
            return await self.first(**kwargs)
        except self.source_class.DoesNotExist:
            pass
        return None

    def filter(self, *args: Any, **kwargs: Any) -> "AsyncBaseSet":
        """
        Apply filters to the existing nodes in the set.

        :param args: a Q object

            e.g `.filter(Q(salary__lt=10000) | Q(salary__gt=20000))`.

        :param kwargs: filter parameters

            Filters mimic Django's syntax with the double '__' to separate field and operators.

            e.g `.filter(salary__gt=20000)` results in `salary > 20000`.

            The following operators are available:

             * 'lt': less than
             * 'gt': greater than
             * 'lte': less than or equal to
             * 'gte': greater than or equal to
             * 'ne': not equal to
             * 'in': matches one of list (or tuple)
             * 'isnull': is null
             * 'regex': matches supplied regex (neo4j regex format)
             * 'exact': exactly match string (just '=')
             * 'iexact': case insensitive match string
             * 'contains': contains string
             * 'icontains': case insensitive contains
             * 'startswith': string starts with
             * 'istartswith': case insensitive string starts with
             * 'endswith': string ends with
             * 'iendswith': case insensitive string ends with

        :return: self
        """
        if args or kwargs:
            # Need to grab and remove the VectorFilter from both args and kwargs
            new_args = [] # As args are a tuple, theyre immutable. But we need to remove the vectorfilter from the arguments so they dont go into Q. 
            for arg in args:
                if isinstance(arg, VectorFilter) and (not self._vector_query):
                    self._vector_query = arg
                new_args.append(arg)

            new_args = tuple(new_args)

            if kwargs.get("vector_filter"):
                if isinstance(kwargs["vector_filter"], VectorFilter) and (not self._vector_query):
                    self._vector_query = kwargs.pop("vector_filter")
                    

            self.q_filters = Q(self.q_filters & Q(*new_args, **kwargs))

        return self

    def exclude(self, *args: Any, **kwargs: Any) -> "BaseSet":
        """
        Exclude nodes from the NodeSet via filters.

        :param kwargs: filter parameters see syntax for the filter method
        :return: self
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & ~Q(*args, **kwargs))
        return self

    def has(self, **kwargs: Any) -> "BaseSet":
        must_match, dont_match = process_has_args(self.source_class, kwargs)
        self.must_match.update(must_match)
        self.dont_match.update(dont_match)
        return self

    def order_by(self, *props: Any) -> "BaseSet":
        """
        Order by properties. Prepend with minus to do descending. Pass None to
        remove ordering.
        """
        should_remove = len(props) == 1 and props[0] is None
        if not hasattr(self, "order_by_elements") or should_remove:
            self.order_by_elements = []
            if should_remove:
                return self
        if "?" in props:
            self.order_by_elements.append("?")
        else:
            for prop in props:
                if isinstance(prop, RawCypher):
                    self.order_by_elements.append(prop)
                    continue
                prop = prop.strip()
                if prop.startswith("-"):
                    prop = prop[1:]
                    desc = True
                else:
                    desc = False

                if prop in self.source_class.defined_properties(rels=False):
                    property_obj = getattr(self.source_class, prop)
                    if isinstance(property_obj, AliasProperty):
                        prop = property_obj.aliased_to()

                self.order_by_elements.append(prop + (" DESC" if desc else ""))

        return self

    def _register_relation_to_fetch(
        self, relation_def: Any, alias: TOptional[str] = None
    ) -> "Path":
        if isinstance(relation_def, Path):
            item = relation_def
        else:
            item = Path(
                value=relation_def,
            )
        if alias:
            item.alias = alias
        return item

    def unique_variables(self, *paths: tuple[str, ...]) -> "NodeSet":
        """Generate unique variable names for the given paths."""
        self._unique_variables = paths
        return self

    def traverse(self, *paths: tuple[str, ...], **aliased_paths: dict) -> "NodeSet":
        """Specify a set of paths to traverse."""
        relations = []
        for path in paths:
            relations.append(self._register_relation_to_fetch(path))
        for alias, aliased_path in aliased_paths.items():
            relations.append(
                self._register_relation_to_fetch(aliased_path, alias=alias)
            )
        self.relations_to_fetch = relations
        return self

    def fetch_relations(self, *relation_names: tuple[str, ...]) -> "NodeSet":
        """Specify a set of relations to traverse and return."""
        warnings.warn(
            "fetch_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )
        relations = []
        for relation_name in relation_names:
            if isinstance(relation_name, Optional):
                relation_name = Path(value=relation_name.relation, optional=True)
            relations.append(self._register_relation_to_fetch(relation_name))
        self.relations_to_fetch = relations
        return self

    def traverse_relations(
        self, *relation_names: tuple[str, ...], **aliased_relation_names: dict
    ) -> "NodeSet":
        """Specify a set of relations to traverse only."""

        warnings.warn(
            "traverse_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )

        def convert_to_path(input: Union[str, Optional]) -> Path:
            self.q_filters = Q(self.q_filters & Q(*args, **kwargs))
        return self

    def exclude(self, *args: Any, **kwargs: Any) -> "AsyncBaseSet":
        """
        Exclude nodes from the NodeSet via filters.

        :param kwargs: filter parameters see syntax for the filter method
        :return: self
        """
        if args or kwargs:
            self.q_filters = Q(self.q_filters & ~Q(*args, **kwargs))
        return self

    def has(self, **kwargs: Any) -> "AsyncBaseSet":
        must_match, dont_match = process_has_args(self.source_class, kwargs)
        self.must_match.update(must_match)
        self.dont_match.update(dont_match)
        return self

    def order_by(self, *props: Any) -> "AsyncBaseSet":
        """
        Order by properties. Prepend with minus to do descending. Pass None to
        remove ordering.
        """
        should_remove = len(props) == 1 and props[0] is None
        if not hasattr(self, "order_by_elements") or should_remove:
            self.order_by_elements = []
            if should_remove:
                return self
        if "?" in props:
            self.order_by_elements.append("?")
        else:
            for prop in props:
                if isinstance(prop, RawCypher):
                    self.order_by_elements.append(prop)
                    continue
                prop = prop.strip()
                if prop.startswith("-"):
                    prop = prop[1:]
                    desc = True
                else:
                    desc = False

                if prop in self.source_class.defined_properties(rels=False):
                    property_obj = getattr(self.source_class, prop)
                    if isinstance(property_obj, AliasProperty):
                        prop = property_obj.aliased_to()

                self.order_by_elements.append(prop + (" DESC" if desc else ""))

        return self

    def _register_relation_to_fetch(
        self, relation_def: Any, alias: TOptional[str] = None
    ) -> "Path":
        if isinstance(relation_def, Path):
            item = relation_def
        else:
            item = Path(
                value=relation_def,
            )
        if alias:
            item.alias = alias
        return item

    def unique_variables(self, *paths: tuple[str, ...]) -> "AsyncNodeSet":
        """Generate unique variable names for the given paths."""
        self._unique_variables = paths
        return self

    def traverse(
        self, *paths: tuple[str, ...], **aliased_paths: dict
    ) -> "AsyncNodeSet":
        """Specify a set of paths to traverse."""
        relations = []
        for path in paths:
            relations.append(self._register_relation_to_fetch(path))
        for alias, aliased_path in aliased_paths.items():
            relations.append(
                self._register_relation_to_fetch(aliased_path, alias=alias)
            )
        self.relations_to_fetch = relations
        return self

    def fetch_relations(self, *relation_names: tuple[str, ...]) -> "AsyncNodeSet":
        """Specify a set of relations to traverse and return."""
        warnings.warn(
            "fetch_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )
        relations = []
        for relation_name in relation_names:
            if isinstance(relation_name, Optional):
                relation_name = Path(value=relation_name.relation, optional=True)
            relations.append(self._register_relation_to_fetch(relation_name))
        self.relations_to_fetch = relations
        return self

    def traverse_relations(
        self, *relation_names: tuple[str, ...], **aliased_relation_names: dict
    ) -> "AsyncNodeSet":
        """Specify a set of relations to traverse only."""

        warnings.warn(
            "traverse_relations() will be deprecated in version 6, use traverse() instead.",
            DeprecationWarning,
        )

        def convert_to_path(input: Union[str, Optional]) -> Path:
            if isinstance(input, Optional):
                path = Path(value=input.relation, optional=True)
            else:
                path = Path(value=input)
            path.include_nodes_in_return = False
            path.include_rels_in_return = False
            return path

        relations = []
        for relation_name in relation_names:
            relations.append(
                self._register_relation_to_fetch(convert_to_path(relation_name))
            )
        for alias, relation_def in aliased_relation_names.items():
            relations.append(
                self._register_relation_to_fetch(
                    convert_to_path(relation_def), alias=alias
                )
            )

        self.relations_to_fetch = relations
        return self

    def annotate(self, *vars: tuple, **aliased_vars: tuple) -> "AsyncNodeSet":
        """Annotate node set results with extra variables."""

        def register_extra_var(
            vardef: Union[AggregatingFunction, ScalarFunction, Any],
            varname: Union[str, None] = None,
        ) -> None:
            if isinstance(vardef, (AggregatingFunction, ScalarFunction)):
                self._extra_results.append(
                    {"vardef": vardef, "alias": varname if varname else ""}
                )
            else:
                raise NotImplementedError

        for vardef in vars:
            register_extra_var(vardef)
        for varname, vardef in aliased_vars.items():
            register_extra_var(vardef, varname)

        return self

    def _to_subgraph(self, root_node: Any, other_nodes: Any, subgraph: dict) -> Any:
        """Recursive method to build root_node's relation graph from subgraph."""
        root_node._relations = {}
        for name, relation_def in subgraph.items():
            for var_name, node in other_nodes.items():
                if (
                    var_name
                    not in [
                        relation_def["variable_name"],
                        relation_def["rel_variable_name"],
                    ]
                    or node is None
                ):
                    continue
                if isinstance(node, list):
                    if len(node) > 0 and isinstance(node[0], AsyncStructuredRel):
                        name += "_relationship"
                    root_node._relations[name] = []
                    for item in node:
                        root_node._relations[name].append(
                            self._to_subgraph(
                                item, other_nodes, relation_def["children"]
                            )
                        )
                else:
                    if isinstance(node, AsyncStructuredRel):
                        name += "_relationship"
                    root_node._relations[name] = self._to_subgraph(
                        node, other_nodes, relation_def["children"]
                    )

        return root_node

    async def resolve_subgraph(self) -> list:
        """
        Convert every result contained in this node set to a subgraph.

        By default, we receive results from neomodel as a list of
        nodes without the hierarchy. This method tries to rebuild this
        hierarchy without overriding anything in the node, that's why
        we use a dedicated property to store node's relations.

        """
        if (
            self.relations_to_fetch
            and not self.relations_to_fetch[0].include_nodes_in_return
            and not self.relations_to_fetch[0].include_rels_in_return
        ):
            raise NotImplementedError(
                "You cannot use traverse_relations() with resolve_subgraph(), use fetch_relations() instead."
            )
        results: list = []
        qbuilder = self.query_cls(self)
        await qbuilder.build_ast()
        if not qbuilder._ast.subgraph:
            raise RuntimeError(
                "Nothing to resolve. Make sure to include relations in the result using fetch_relations() or filter()."
            )
        other_nodes = {}
        root_node = None
        async for row in qbuilder._execute(dict_output=True):
            for name, node in row.items():
                if node.__class__ is self.source and "_" not in name:
                    root_node = node
                    continue
                if isinstance(node, list) and isinstance(node[0], list):
                    other_nodes[name] = node[0]
                    continue
                other_nodes[name] = node
            results.append(
                self._to_subgraph(root_node, other_nodes, qbuilder._ast.subgraph)
            )
        return results

    async def subquery(
        self,
        nodeset: "AsyncNodeSet",
        return_set: list[str],
        initial_context: TOptional[list[str]] = None,
    ) -> "AsyncNodeSet":
        """Add a subquery to this node set.

        A subquery is a regular cypher query but executed within the context of a CALL
        statement. Such query will generally fetch additional variables which must be
        declared inside return_set variable in order to be included in the final RETURN
        statement.
        """
        namespace = f"sq{len(self._subqueries) + 1}"
        qbuilder = await nodeset.query_cls(
            nodeset, subquery_namespace=namespace
        ).build_ast()
        for var in return_set:
            if (
                var != qbuilder._ast.return_clause
                and (
                    not qbuilder._ast.additional_return
                    or var not in qbuilder._ast.additional_return
                )
                and var
                not in [res["alias"] for res in nodeset._extra_results if res["alias"]]
                and var
                not in [
                    varname
                    for tr in nodeset._intermediate_transforms
                    for varname, vardef in tr["vars"].items()
                    if vardef.get("include_in_return")
                ]
            ):
                raise RuntimeError(f"Variable '{var}' is not returned by subquery.")
        if initial_context:
            for var in initial_context:
                if type(var) is not str and not isinstance(
                    var, (NodeNameResolver, RelationNameResolver, RawCypher)
                ):
                    raise ValueError(
                        f"Wrong variable specified in initial context, should be a string or an instance of NodeNameResolver or RelationNameResolver"
                    )
        self._subqueries.append(
            {
                "query": qbuilder.build_query(),
                "query_params": qbuilder._query_params,
                "return_set": return_set,
                "initial_context": initial_context,
            }
        )
        return self

    def intermediate_transform(
        self,
        vars: dict[str, Transformation],
        distinct: bool = False,
        ordering: TOptional[list] = None,
    ) -> "AsyncNodeSet":
        if not vars:
            raise ValueError(
                "You must provide one variable at least when calling intermediate_transform()"
            )
        for name, props in vars.items():
            source = props["source"]
            if type(source) is not str and not isinstance(
                source, (NodeNameResolver, RelationNameResolver, RawCypher)
            ):
                raise ValueError(
                    f"Wrong source type specified for variable '{name}', should be a string or an instance of NodeNameResolver or RelationNameResolver"
                )
        self._intermediate_transforms.append(
            {"vars": vars, "distinct": distinct, "ordering": ordering}
        )
        return self


class AsyncTraversal(AsyncBaseSet):
    """
    Models a traversal from a node to another.

    :param source: Starting of the traversal.
    :type source: A :class:`~neomodel.core.StructuredNode` subclass, an
                  instance of such, a :class:`~neomodel.match.NodeSet` instance
                  or a :class:`~neomodel.match.Traversal` instance.
    :param name: A name for the traversal.
    :type name: :class:`str`
    :param definition: A relationship definition that most certainly deserves
                       a documentation here.
    :type definition: :class:`dict`
    """

    definition: dict
    source: Any
    source_class: Any
    target_class: Any
    name: str
    filters: list

    def __await__(self) -> Any:
        return self.all().__await__()  # type: ignore[attr-defined]

    def __init__(self, source: Any, name: str, definition: dict) -> None:
        """
        Create a traversal

        """
        self.source = source

        if isinstance(source, AsyncTraversal):
            self.source_class = source.target_class
        elif inspect.isclass(source) and issubclass(source, AsyncStructuredNode):
            self.source_class = source
        elif isinstance(source, AsyncStructuredNode):
            self.source_class = source.__class__
        elif isinstance(source, AsyncNodeSet):
            self.source_class = source.source_class
        else:
            raise TypeError(f"Bad source for traversal: {type(source)}")

        invalid_keys = set(definition) - {
            "direction",
            "model",
            "node_class",
            "relation_type",
        }
        if invalid_keys:
            raise ValueError(f"Prohibited keys in Traversal definition: {invalid_keys}")

        self.definition = definition
        self.target_class = definition["node_class"]
        self.name = name
        self.filters: list = []

    def match(self, **kwargs: Any) -> "AsyncTraversal":
        """
        Traverse relationships with properties matching the given parameters.

            e.g: `.match(price__lt=10)`

        :param kwargs: see `NodeSet.filter()` for syntax
        :return: self
        """
        if kwargs:
            if self.definition.get("model") is None:
                raise ValueError(
                    "match() with filter only available on relationships with a model"
                )
            output = process_filter_args(self.definition["model"], kwargs)
            if output:
                self.filters.append(output)
        return self



from test._async_compat import mark_sync_test

from neomodel import (
    IntegerProperty,
    NeomodelPath,
    RelationshipTo,
    StringProperty,
    StructuredNode,
    StructuredRel,
    UniqueIdProperty,
    db,
)


class PersonLivesInCity(StructuredRel):
    """
    Relationship with data that will be instantiated as "stand-alone"
    """

    some_num = IntegerProperty(index=True, default=12)


class CountryOfOrigin(StructuredNode):
    code = StringProperty(unique_index=True, required=True)


class CityOfResidence(StructuredNode):
    name = StringProperty(required=True)
    country = RelationshipTo(CountryOfOrigin, "FROM_COUNTRY")


class PersonOfInterest(StructuredNode):
    uid = UniqueIdProperty()
    name = StringProperty(unique_index=True)
    age = IntegerProperty(index=True, default=0)

    country = RelationshipTo(CountryOfOrigin, "IS_FROM")
    city = RelationshipTo(CityOfResidence, "LIVES_IN", model=PersonLivesInCity)


@mark_sync_test
def test_path_instantiation():
    """
    Neo4j driver paths should be instantiated as neomodel paths, with all of
    their nodes and relationships resolved to their Python objects wherever
    such a mapping is available.
    """

    c1 = CountryOfOrigin(code="GR").save()
    c2 = CountryOfOrigin(code="FR").save()

    ct1 = CityOfResidence(name="Athens", country=c1).save()
    ct2 = CityOfResidence(name="Paris", country=c2).save()

    p1 = PersonOfInterest(name="Bill", age=22).save()
    p1.country.connect(c1)
    p1.city.connect(ct1)

    p2 = PersonOfInterest(name="Jean", age=28).save()
    p2.country.connect(c2)
    p2.city.connect(ct2)

    p3 = PersonOfInterest(name="Bo", age=32).save()
    p3.country.connect(c1)
    p3.city.connect(ct2)

    p4 = PersonOfInterest(name="Drop", age=16).save()
    p4.country.connect(c1)
    p4.city.connect(ct2)

    # Retrieve a single path
    q = db.cypher_query(
        "MATCH p=(:CityOfResidence{name:'Athens'})<-[:LIVES_IN]-(:PersonOfInterest)-[:IS_FROM]->(:CountryOfOrigin) RETURN p LIMIT 1",
        resolve_objects=True,
    )

    path_object = q[0][0][0]
    path_nodes = path_object.nodes
    path_rels = path_object.relationships

    assert isinstance(path_object, NeomodelPath)
    assert isinstance(path_nodes[0], CityOfResidence)
    assert isinstance(path_nodes[1], PersonOfInterest)
    assert isinstance(path_nodes[2], CountryOfOrigin)
    assert isinstance(path_object.start_node, CityOfResidence)
    assert isinstance(path_object.end_node, CountryOfOrigin)

    assert isinstance(path_rels[0], PersonLivesInCity)
    assert isinstance(path_rels[1], StructuredRel)

    path_string = str(path_object)
    assert path_string.startswith("<Path start=<CityOfResidence")
    assert path_string.endswith("size=2>")
    assert len(path_object) == 2
    for rel in path_object:
        assert isinstance(rel, StructuredRel)

    c1.delete()
    c2.delete()
    ct1.delete()
    ct2.delete()
    p1.delete()
    p2.delete()
    p3.delete()
    p4.delete()



import pytest
import pytest_asyncio

mark_async_test = pytest.mark.asyncio(loop_scope="session")
mark_async_session_auto_fixture = pytest_asyncio.fixture(
    loop_scope="session", scope="session", autouse=True
)
mark_async_function_auto_fixture = pytest_asyncio.fixture(
    loop_scope="session", autouse=True
)
mark_sync_session_auto_fixture = pytest.fixture(scope="session", autouse=True)
mark_sync_function_auto_fixture = pytest.fixture(autouse=True)


def mark_sync_test(f):
    return f


class AsyncTestDecorators:
    mark_async_only_test = mark_async_test


class TestDecorators:
    @staticmethod
    def mark_async_only_test(f):
        skip_decorator = pytest.mark.skip("Async only test")
        return skip_decorator(f)



import types
from typing import Any

from neo4j.graph import Entity

from neomodel.exceptions import RequiredProperty
from neomodel.properties import AliasProperty, Property


def display_for(key: str) -> Any:
    def display_choice(self: Any) -> Any:
        return getattr(self.__class__, key).choices[getattr(self, key)]

    return display_choice


class AsyncPropertyManager:
    """
    Common methods for handling properties on node and relationship objects.
    """

    def __init__(self, **kwargs: dict[str, Any]) -> None:
        properties = getattr(self, "__all_properties__", None)
        if properties is None:
            properties = self.defined_properties(rels=False, aliases=False).items()
        for name, property in properties:
            if kwargs.get(name) is None:
                if getattr(property, "has_default", False):
                    setattr(self, name, property.default_value())
                else:
                    setattr(self, name, None)
            else:
                setattr(self, name, kwargs[name])

            if getattr(property, "choices", None):
                setattr(
                    self,
                    f"get_{name}_display",
                    types.MethodType(display_for(name), self),
                )

            if name in kwargs:
                del kwargs[name]

        aliases = getattr(self, "__all_aliases__", None)
        if aliases is None:
            aliases = self.defined_properties(
                aliases=True, rels=False, properties=False
            ).items()
        for name, property in aliases:
            if name in kwargs:
                setattr(self, name, kwargs[name])
                del kwargs[name]

        # undefined properties (for magic @prop.setters etc)
        for name, property in kwargs.items():
            setattr(self, name, property)

    @property
    def __properties__(self) -> dict[str, Any]:
        from neomodel.async_.relationship_manager import AsyncRelationshipManager

        return dict(
            (name, value)
            for name, value in vars(self).items()
            if not name.startswith("_")
            and not callable(value)
            and not isinstance(
                value,
                (
                    AsyncRelationshipManager,
                    AliasProperty,
                ),
            )
        )

    @classmethod
    def deflate(
        cls, properties: Any, obj: Any = None, skip_empty: bool = False
    ) -> dict[str, Any]:
        """
        Deflate the properties of a PropertyManager subclass (a user-defined StructuredNode or StructuredRel) so that it
        can be put into a neo4j.graph.Entity (a neo4j.graph.Node or neo4j.graph.Relationship) for storage. properties
        can be constructed manually, or fetched from a PropertyManager subclass using __properties__.

        Includes mapping from python class attribute name -> database property name (see Property.db_property).

        Ignores any properties that are not defined as python attributes in the class definition.
        """
        deflated = {}
        for name, property in cls.defined_properties(aliases=False, rels=False).items():
            db_property = property.get_db_property_name(name)
            if properties.get(name) is not None:
                deflated[db_property] = property.deflate(properties[name], obj)
            elif property.has_default:
                deflated[db_property] = property.deflate(property.default_value(), obj)
            elif property.required:
                raise RequiredProperty(name, cls)
            elif not skip_empty:
                deflated[db_property] = None
        return deflated

    @classmethod
    def inflate(cls: Any, graph_entity: Entity) -> Any:
        """
        Inflate the properties of a neo4j.graph.Entity (a neo4j.graph.Node or neo4j.graph.Relationship) into an instance
        of cls.
        Includes mapping from database property name (see Property.db_property) -> python class attribute name.
        Ignores any properties that are not defined as python attributes in the class definition.
        """
        inflated = {}
        for name, property in cls.defined_properties(aliases=False, rels=False).items():
            db_property = property.get_db_property_name(name)
            if db_property in graph_entity:
                inflated[name] = property.inflate(
                    graph_entity[db_property], graph_entity
                )
            elif property.has_default:
                inflated[name] = property.default_value()
            else:
                inflated[name] = None
        return cls(**inflated)

    @classmethod
    def defined_properties(
        cls: Any, aliases: bool = True, properties: bool = True, rels: bool = True
    ) -> dict[str, Any]:
        from neomodel.async_.relationship_manager import AsyncRelationshipDefinition

        props = {}
        for baseclass in reversed(cls.__mro__):
            props.update(
                dict(
                    (name, property)
                    for name, property in vars(baseclass).items()
                    if (aliases and isinstance(property, AliasProperty))
                    or (
                        properties
                        and isinstance(property, Property)
                        and not isinstance(property, AliasProperty)
                    )
                    or (rels and isinstance(property, AsyncRelationshipDefinition))
                )
            )
        return props



import os
import warnings
from test._async_compat import (
    mark_async_function_auto_fixture,
    mark_sync_session_auto_fixture,
)

from neomodel import config, db


@mark_sync_session_auto_fixture
def setup_neo4j_session(request):
    """
    Provides initial connection to the database and sets up the rest of the test suite

    :param request: The request object. Please see <https://docs.pytest.org/en/latest/reference.html#_pytest.hookspec.pytest_sessionstart>`_
    :type Request object: For more information please see <https://docs.pytest.org/en/latest/reference.html#request>`_
    """

    warnings.simplefilter("default")

    config.DATABASE_URL = os.environ.get(
        "NEO4J_BOLT_URL", "bolt://neo4j:foobarbaz@localhost:7687"
    )

    # Clear the database if required
    database_is_populated, _ = db.cypher_query(
        "MATCH (a) return count(a)>0 as database_is_populated"
    )
    if database_is_populated[0][0] and not request.config.getoption("resetdb"):
        raise SystemError(
            "Please note: The database seems to be populated.\n\tEither delete all nodes and edges manually, or set the --resetdb parameter when calling pytest\n\n\tpytest --resetdb."
        )

    db.clear_neo4j_database(clear_constraints=True, clear_indexes=True)

    db.install_all_labels()

    db.cypher_query(
        "CREATE OR REPLACE USER troygreene SET PASSWORD 'foobarbaz' CHANGE NOT REQUIRED"
    )
    db_edition = db.database_edition
    if db_edition == "enterprise":
        db.cypher_query("GRANT ROLE publisher TO troygreene")
        db.cypher_query("GRANT IMPERSONATE (troygreene) ON DBMS TO admin")

    yield

    db.close_connection()


@mark_async_function_auto_fixture
def setUp():
    db.cypher_query("MATCH (n) DETACH DELETE n")
    yield



from test._async_compat import mark_async_test

import pytest
from pytest import raises

from neomodel import (
    AsyncStructuredNode,
    IntegerProperty,
    StringProperty,
    UniqueProperty,
    adb,
)
from neomodel.exceptions import ConstraintValidationFailed


class Human(AsyncStructuredNode):
    name = StringProperty(unique_index=True)
    age = IntegerProperty(index=True)


@mark_async_test
async def test_unique_error():
    await adb.install_labels(Human)
    await Human(name="j1m", age=13).save()
    try:
        await Human(name="j1m", age=14).save()
    except UniqueProperty as e:
        assert str(e).find("j1m")
        assert str(e).find("name")
    else:
        assert False, "UniqueProperty not raised."


@mark_async_test
async def test_existence_constraint_error():
    if not await adb.edition_is_enterprise():
        pytest.skip("Skipping test for community edition")
    await adb.cypher_query(
        "CREATE CONSTRAINT test_existence_constraint FOR (n:Human) REQUIRE n.age IS NOT NULL"
    )
    with raises(ConstraintValidationFailed, match=r"must have the property"):
        await Human(name="Scarlett").save()

    await adb.cypher_query("DROP CONSTRAINT test_existence_constraint")


@mark_async_test
async def test_optional_properties_dont_get_indexed():
    await Human(name="99", age=99).save()
    h = await Human.nodes.get(age=99)
    assert h
    assert h.name == "99"

    await Human(name="98", age=98).save()
    h = await Human.nodes.get(age=98)
    assert h
    assert h.name == "98"


@mark_async_test
async def test_escaped_chars():
    _name = "sarah:test"
    await Human(name=_name, age=3).save()
    r = await Human.nodes.filter(name=_name)
    assert r[0].name == _name


@mark_async_test
async def test_does_not_exist():
    with raises(Human.DoesNotExist):
        await Human.nodes.get(name="XXXX")


@mark_async_test
async def test_custom_label_name():
    class Giraffe(AsyncStructuredNode):
        __label__ = "Giraffes"
        name = StringProperty(unique_index=True)

    jim = await Giraffe(name="timothy").save()
    node = await Giraffe.nodes.get(name="timothy")
    assert node.name == jim.name

    class SpecialGiraffe(Giraffe):
        power = StringProperty()

    # custom labels aren't inherited
    assert SpecialGiraffe.__label__ == "SpecialGiraffe"



from test._async_compat import mark_async_test

from neomodel import AsyncRelationshipTo, AsyncStructuredNode


class SomeModel(AsyncStructuredNode):
    test = AsyncRelationshipTo("SomeModel", "SELF")


@mark_async_test
async def test_len_relationship():
    t1 = await SomeModel().save()
    t2 = await SomeModel().save()

    await t1.test.connect(t2)
    l = len(await t1.test.all())

    assert l
    assert l == 1



import sys
from importlib import import_module
from os import path


def load_python_module_or_file(name):
    """
    Imports an existing python module or file into the current workspace.

    In both cases, *the resource must exist*.

    :param name: A string that refers either to a Python module or a source coe
                 file to load in the current workspace.
    :type name: str
    """
    # Is a file
    if name.lower().endswith(".py"):
        basedir = path.dirname(path.abspath(name))
        # Add base directory to pythonpath
        sys.path.append(basedir)
        module_name = path.basename(name)[:-3]

    else:  # A module
        # Add current directory to pythonpath
        sys.path.append(path.abspath(path.curdir))

        module_name = name

    if module_name.startswith("."):
        pkg = module_name.split(".")[1]
    else:
        pkg = None

    import_module(module_name, package=pkg)
    print(f"Loaded {name}")


def recursive_list_classes(cls, exclude_list=None):  # recursively return all subclasses
    subclasses = cls.__subclasses__()
    if not subclasses:  # base case: no more subclasses
        return []
    elif cls not in exclude_list:
        return [s for s in subclasses if s not in exclude_list] + [
            g
            for s in cls.__subclasses__()
            for g in recursive_list_classes(s, exclude_list=exclude_list)
        ]



import os
from test._async_compat import mark_sync_test
from test.conftest import NEO4J_PASSWORD, NEO4J_URL, NEO4J_USERNAME

import pytest
from neo4j import Driver, GraphDatabase
from neo4j.debug import watch

from neomodel import StringProperty, StructuredNode, config, db


@mark_sync_test
@pytest.fixture(autouse=True)
def setup_teardown():
    yield
    # Teardown actions after tests have run
    # Reconnect to initial URL for potential subsequent tests
    db.close_connection()
    db.set_connection(url=config.DATABASE_URL)


@pytest.fixture(autouse=True, scope="session")
def neo4j_logging():
    with watch("neo4j"):
        yield


@mark_sync_test
def get_current_database_name() -> str:
    """
    Fetches the name of the currently active database from the Neo4j database.

    Returns:
    - str: The name of the current database.
    """
    results, meta = db.cypher_query("CALL db.info")
    results_as_dict = [dict(zip(meta, row)) for row in results]

    return results_as_dict[0]["name"]


class Pastry(StructuredNode):
    name = StringProperty(unique_index=True)


@mark_sync_test
def test_set_connection_driver_works():
    # Verify that current connection is up
    assert Pastry(name="Chocolatine").save()
    db.close_connection()

    # Test connection using a driver
    db.set_connection(
        driver=GraphDatabase().driver(NEO4J_URL, auth=(NEO4J_USERNAME, NEO4J_PASSWORD))
    )
    assert Pastry(name="Croissant").save()


@mark_sync_test
def test_config_driver_works():
    # Verify that current connection is up
    assert Pastry(name="Chausson aux pommes").save()
    db.close_connection()

    # Test connection using a driver defined in config
    driver: Driver = GraphDatabase().driver(
        NEO4J_URL, auth=(NEO4J_USERNAME, NEO4J_PASSWORD)
    )

    config.DRIVER = driver
    assert Pastry(name="Grignette").save()

    # Clear config
    # No need to close connection - pytest teardown will do it
    config.DRIVER = None


@mark_sync_test
def test_connect_to_non_default_database():
    if not db.edition_is_enterprise():
        pytest.skip("Skipping test for community edition - no multi database in CE")
    database_name = "pastries"
    db.cypher_query(f"CREATE DATABASE {database_name} IF NOT EXISTS")
    db.close_connection()

    # Set database name in url - for url init only
    db.set_connection(url=f"{config.DATABASE_URL}/{database_name}")
    assert get_current_database_name() == "pastries"

    db.close_connection()

    # Set database name in config - for both url and driver init
    config.DATABASE_NAME = database_name

    # url init
    db.set_connection(url=config.DATABASE_URL)
    assert get_current_database_name() == "pastries"

    db.close_connection()

    # driver init
    db.set_connection(
        driver=GraphDatabase().driver(NEO4J_URL, auth=(NEO4J_USERNAME, NEO4J_PASSWORD))
    )
    assert get_current_database_name() == "pastries"

    # Clear config
    # No need to close connection - pytest teardown will do it
    config.DATABASE_NAME = None


@mark_sync_test
@pytest.mark.parametrize(
    "url", ["bolt://user:password", "http://user:password@localhost:7687"]
)
def test_wrong_url_format(url):
    with pytest.raises(
        ValueError,
        match=rf"Expecting url format: bolt://user:password@localhost:7687 got {url}",
    ):
        db.set_connection(url=url)


@mark_sync_test
@pytest.mark.parametrize("protocol", ["neo4j+s", "neo4j+ssc", "bolt+s", "bolt+ssc"])
def test_connect_to_aura(protocol):
    # Skip test if Aura credentials are not available (e.g., in external PRs)
    required_env_vars = [
        "AURA_TEST_DB_USER",
        "AURA_TEST_DB_PASSWORD",
        "AURA_TEST_DB_HOSTNAME",
    ]
    missing_vars = [var for var in required_env_vars if var not in os.environ]
    if missing_vars:
        pytest.skip(
            f"Skipping Aura test - missing environment variables: {', '.join(missing_vars)}"
        )

    cypher_return = "hello world"
    default_cypher_query = f"RETURN '{cypher_return}'"
    db.close_connection()

    _set_connection(protocol=protocol)
    result, _ = db.cypher_query(default_cypher_query)

    assert len(result) > 0
    assert result[0][0] == cypher_return


def _set_connection(protocol):
    AURA_TEST_DB_USER = os.environ["AURA_TEST_DB_USER"]
    AURA_TEST_DB_PASSWORD = os.environ["AURA_TEST_DB_PASSWORD"]
    AURA_TEST_DB_HOSTNAME = os.environ["AURA_TEST_DB_HOSTNAME"]

    database_url = f"{protocol}://{AURA_TEST_DB_USER}:{AURA_TEST_DB_PASSWORD}@{AURA_TEST_DB_HOSTNAME}"
    db.set_connection(url=database_url)



import functools
import json
import re
import uuid
from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from typing import Any, Callable, Optional

import neo4j.time
import pytz

from neomodel import config
from neomodel.exceptions import DeflateError, InflateError, NeomodelException

TOO_MANY_DEFAULTS = "too many defaults"


def validator(fn: Callable) -> Callable:
    fn_name = fn.func_name if hasattr(fn, "func_name") else fn.__name__
    if fn_name not in ["inflate", "deflate"]:
        raise ValueError("Unknown Property method " + fn_name)

    @functools.wraps(fn)
    def _validator(  # type: ignore
        self, value: Any, obj: Optional[Any] = None, rethrow: Optional[bool] = True
    ) -> Any:
        if rethrow:
            try:
                return fn(self, value)
            except Exception as e:
                if fn_name == "inflate":
                    raise InflateError(self.name, self.owner, str(e), obj) from e
                elif fn_name == "deflate":
                    raise DeflateError(self.name, self.owner, str(e), obj) from e
                else:
                    raise NeomodelException("Unknown Property method " + fn_name) from e
        else:
            # For using with ArrayProperty where we don't want an Inflate/Deflate error.
            return fn(self, value)

    return _validator


class FulltextIndex:
    """
    Fulltext index definition
    """

    def __init__(
        self,
        analyzer: Optional[str] = "standard-no-stop-words",
        eventually_consistent: Optional[bool] = False,
    ):
        """
        Initializes new fulltext index definition with analyzer and eventually consistent

        :param str analyzer: The analyzer to use. Defaults to "standard-no-stop-words".
        :param bool eventually_consistent: Whether the index should be eventually consistent. Defaults to False.
        """
        self.analyzer = analyzer
        self.eventually_consistent = eventually_consistent


class VectorIndex:
    """
    Vector index definition
    """

    def __init__(
        self,
        dimensions: Optional[int] = 1536,
        similarity_function: Optional[str] = "cosine",
    ):
        """
        Initializes new vector index definition with dimensions and similarity

        :param int dimensions: The number of dimensions of the vector. Defaults to 1536.
        :param str similarity_function: The similarity algorithm to use. Defaults to "cosine".
        """
        self.dimensions = dimensions
        self.similarity_function = similarity_function


class Property(metaclass=ABCMeta):
    """
    Base class for object properties.

    :param unique_index: Creates a unique index for this property. Defaults to
                         ``False``.
    :type unique_index: :class:`bool`
    :param index: Creates an index for this property. Defaults to ``False``.
    :type index: :class:`bool`
    :param fulltext_index: Creates a fulltext index for this property. Defaults to ``None``.
    :type fulltext_index: :class:`FulltextIndex`
    :param vector_index: Creates a vector index for this property. Defaults to ``None``.
    :type vector_index: :class:`VectorIndex`
    :param required: Marks the property as required. Defaults to ``False``.
    :type required: :class:`bool`
    :param default: A default value or callable that returns one to set when a
                    node is initialized without specifying this property.
    :param db_property: A name that this property maps to in the database.
                        Defaults to the model's property name.
    :type db_property: :class:`str`
    :param label: Optional, used by ``django_neomodel``.
    :type label: :class:`str`
    :param help_text: Optional, used by ``django_neomodel``.
    :type help_text: :class:`str`
    """

    form_field_class = "CharField"
    name: Optional[str] = None
    owner: Optional[Any] = None
    unique_index: bool = False
    index: bool = False
    fulltext_index: Optional[FulltextIndex] = None
    vector_index: Optional[VectorIndex] = None
    required: bool = False
    default: Any = None
    db_property: Optional[str] = None
    label: Optional[str] = None
    help_text: Optional[str] = None

    # pylint:disable=unused-argument
    def __init__(
        self,
        name: Optional[str] = None,
        owner: Optional[Any] = None,
        unique_index: bool = False,
        index: bool = False,
        fulltext_index: Optional[FulltextIndex] = None,
        vector_index: Optional[VectorIndex] = None,
        required: bool = False,
        default: Optional[Any] = None,
        db_property: Optional[str] = None,
        label: Optional[str] = None,
        help_text: Optional[str] = None,
        **kwargs: dict[str, Any],
    ):
        if default is not None and required:
            raise ValueError(
                "The arguments `required` and `default` are mutually exclusive."
            )
        if unique_index and index:
            raise ValueError(
                "The arguments `unique_index` and `index` are mutually exclusive."
            )

        self.required = required
        self.unique_index = unique_index
        self.index = index
        self.fulltext_index = fulltext_index
        self.vector_index = vector_index
        self.default = default
        self.has_default = self.default is not None
        self.db_property = db_property
        self.label = label
        self.help_text = help_text

    def default_value(self) -> Any:
        """
        Generate a default value

        :return: the value
        """
        if self.has_default:
            if hasattr(self.default, "__call__"):
                return self.default()
            return self.default
        raise ValueError("No default value specified")

    def get_db_property_name(self, attribute_name: str) -> str:
        """
        Returns the name that should be used for the property in the database. This is db_property if supplied upon
        construction, otherwise the given attribute_name from the model is used.
        """
        return self.db_property or attribute_name

    @property
    def is_indexed(self) -> bool:
        return self.unique_index or self.index

    @abstractmethod
    def inflate(self, value: Any, rethrow: bool = False) -> Any:
        pass

    @abstractmethod
    def deflate(self, value: Any, rethrow: bool = False) -> Any:
        pass


class NormalizedProperty(Property):
    """
    Base class for normalized properties. These use the same normalization
    method to in- or deflating.
    """

    @validator
    def inflate(self, value: Any) -> Any:
        return self.normalize(value)

    @validator
    def deflate(self, value: Any) -> Any:
        return self.normalize(value)

    def default_value(self) -> Any:
        default = super().default_value()
        return self.normalize(default)

    def normalize(self, value: Any) -> Any:
        raise NotImplementedError("Specialize normalize method")


class RegexProperty(NormalizedProperty):
    r"""
    Validates a property against a regular expression.

    If sub-classing set:

        expression = r'[^@]+@[^@]+\.[^@]+'
    """

    form_field_class = "RegexField"

    expression: str

    def __init__(self, expression: Optional[str] = None, **kwargs: Any):
        """
        Initializes new property with an expression.

        :param str expression: regular expression validating this property
        """
        super().__init__(**kwargs)
        self.expression = expression or self.expression

    def normalize(self, value: Any) -> str:
        normal = str(value)
        if not re.match(self.expression, normal):
            raise ValueError(f"{value!r} does not match {self.expression!r}")
        return normal


class EmailProperty(RegexProperty):
    """
    Store email addresses
    """

    form_field_class = "EmailField"
    expression = r"[^@]+@[^@]+\.[^@]+"


class StringProperty(NormalizedProperty):
    """
    Stores a unicode string

    :param choices: A mapping of valid strings to label strings that are used
                    to display information in an application. If the default
                    value ``None`` is used, any string is valid.
    :type choices: Any type that can be used to initiate a :class:`dict`.
    :param max_length: The maximum non-zero length that this attribute can be
    :type max_length: int
    """

    def __init__(
        self,
        choices: Optional[Any] = None,
        max_length: Optional[int] = None,
        **kwargs: Any,
    ):
        if max_length is not None:
            if choices is not None:
                raise ValueError(
                    "The arguments `choices` and `max_length` are mutually exclusive."
                )
            if max_length < 1:
                raise ValueError("`max_length` cannot be zero or take negative values.")

        super().__init__(**kwargs)

        self.max_length = max_length
        if choices is None:
            self.choices = None
        else:
            try:
                self.choices = dict(choices)
            except Exception as exc:
                raise ValueError(
                    "The choices argument must be convertible to a dictionary."
                ) from exc
            self.form_field_class = "TypedChoiceField"

    def normalize(self, value: str) -> str:
        # One thing to note here is that the following two checks can remain uncoupled
        # as long as it is guaranteed (by the constructor) that `choices` and `max_length`
        # are mutually exclusive. If that check in the constructor ever has to be removed,
        # these two validation checks here will have to be coupled so that having set
        # `choices` overrides having set the `max_length`.
        if self.choices is not None and value not in self.choices:
            raise ValueError(f"Invalid choice: {value}")
        if self.max_length is not None and len(value) > self.max_length:
            raise ValueError(
                f"Property max length exceeded. Expected {self.max_length}, got {len(value)} == len('{value}')"
            )
        return str(value)

    def default_value(self) -> str:
        return self.normalize(super().default_value())


class IntegerProperty(Property):
    """
    Stores an Integer value
    """

    form_field_class = "IntegerField"

    @validator
    def inflate(self, value: Any) -> int:
        return int(value)

    @validator
    def deflate(self, value: Any) -> int:
        return int(value)

    def default_value(self) -> int:
        return int(super().default_value())


class ArrayProperty(Property):
    """
    Stores a list of items
    """

    def __init__(self, base_property: Optional[Property] = None, **kwargs: Any):
        """
        Store a list of values, optionally of a specific type.

        :param base_property: List item type e.g StringProperty for string
        :type: Property
        """

        # list item type
        if base_property is not None:
            if not isinstance(base_property, Property):
                raise TypeError("Expecting neomodel Property")

            if isinstance(base_property, ArrayProperty):
                raise TypeError("Cannot have nested ArrayProperty")

            for illegal_attr in [
                "default",
                "index",
                "unique_index",
                "required",
            ]:
                if getattr(base_property, illegal_attr, None):
                    raise ValueError(
                        f'ArrayProperty base_property cannot have "{illegal_attr}" set'
                    )

        self.base_property = base_property

        super().__init__(**kwargs)

    @validator
    def inflate(self, value: Any) -> list:
        if self.base_property:
            return [self.base_property.inflate(item, rethrow=False) for item in value]

        return list(value)

    @validator
    def deflate(self, value: Any) -> list:
        if self.base_property:
            return [self.base_property.deflate(item, rethrow=False) for item in value]

        return list(value)

    def default_value(self) -> list:
        return list(super().default_value())


class FloatProperty(Property):
    """
    Store a floating point value
    """

    form_field_class = "FloatField"

    @validator
    def inflate(self, value: Any) -> float:
        return float(value)

    @validator
    def deflate(self, value: Any) -> float:
        return float(value)

    def default_value(self) -> float:
        return float(super().default_value())


class BooleanProperty(Property):
    """
    Stores a boolean value
    """

    form_field_class = "BooleanField"

    @validator
    def inflate(self, value: Any) -> bool:
        return bool(value)

    @validator
    def deflate(self, value: Any) -> bool:
        return bool(value)

    def default_value(self) -> bool:
        return bool(super().default_value())


class DateProperty(Property):
    """
    Stores a date
    """

    form_field_class = "DateField"

    @validator
    def inflate(self, value: Any) -> date:
        if isinstance(value, neo4j.time.DateTime):
            value = date(value.year, value.month, value.day)
        elif isinstance(value, str) and "T" in value:
            value = value[: value.find("T")]
        return datetime.strptime(str(value), "%Y-%m-%d").date()

    @validator
    def deflate(self, value: date) -> str:
        if not isinstance(value, date):
            msg = f"datetime.date object expected, got {repr(value)}"
            raise ValueError(msg)
        return value.isoformat()


class DateTimeFormatProperty(Property):
    """
    Store a datetime by custom format
    :param default_now: If ``True``, the creation time (Local) will be used as default.
    Defaults to ``False``.
    :param format:      Date format string, default is %Y-%m-%d

    :type default_now:  :class:`bool`
    :type format:       :class:`str`
    """

    form_field_class = "DateTimeFormatField"

    def __init__(
        self, default_now: bool = False, format: str = "%Y-%m-%d", **kwargs: Any
    ):
        if default_now:
            if "default" in kwargs:
                raise ValueError(TOO_MANY_DEFAULTS)
            kwargs["default"] = datetime.now()

        self.format = format
        super().__init__(**kwargs)

    @validator
    def inflate(self, value: Any) -> datetime:
        return datetime.strptime(str(value), self.format)

    @validator
    def deflate(self, value: datetime) -> str:
        if not isinstance(value, datetime):
            raise ValueError(f"datetime object expected, got {type(value)}.")
        return datetime.strftime(value, self.format)


class DateTimeProperty(Property):
    """A property representing a :class:`datetime.datetime` object as
    unix epoch.

    :param default_now: If ``True``, the creation time (UTC) will be used as default.
                        Defaults to ``False``.
    :type default_now: :class:`bool`
    """

    form_field_class = "DateTimeField"

    def __init__(self, default_now: bool = False, **kwargs: Any):
        if default_now:
            if "default" in kwargs:
                raise ValueError(TOO_MANY_DEFAULTS)
            kwargs["default"] = lambda: datetime.now(pytz.utc)

        super().__init__(**kwargs)

    @validator
    def inflate(self, value: Any) -> datetime:
        try:
            epoch = float(value)
        except ValueError as exc:
            raise ValueError(
                f"Float or integer expected, got {type(value)} cannot inflate to datetime."
            ) from exc
        except TypeError as exc:
            raise TypeError(
                f"Float or integer expected. Can't inflate {type(value)} to datetime."
            ) from exc
        return datetime.fromtimestamp(epoch, tz=pytz.utc)

    @validator
    def deflate(self, value: datetime) -> float:
        if not isinstance(value, datetime):
            raise ValueError(f"datetime object expected, got {type(value)}.")
        if value.tzinfo:
            value = value.astimezone(pytz.utc)
            epoch_date = datetime(1970, 1, 1, tzinfo=pytz.utc)
        elif config.FORCE_TIMEZONE:
            raise ValueError(f"Error deflating {value}: No timezone provided.")
        else:
            # No timezone specified on datetime object.. assuming UTC
            epoch_date = datetime(1970, 1, 1)
        return float((value - epoch_date).total_seconds())


class DateTimeNeo4jFormatProperty(Property):
    """
    Store a datetime by native neo4j format

    :param default_now: If ``True``, the creation time (Local) will be used as default.
                        Defaults to ``False``.

    :type default_now:  :class:`bool`
    """

    form_field_class = "DateTimeNeo4jFormatField"

    def __init__(self, default_now: bool = False, **kwargs: Any):
        if default_now:
            if "default" in kwargs:
                raise ValueError(TOO_MANY_DEFAULTS)
            kwargs["default"] = datetime.now()

        self.format = format
        super(DateTimeNeo4jFormatProperty, self).__init__(**kwargs)

    @validator
    def inflate(self, value: Any) -> datetime:
        return value.to_native()

    @validator
    def deflate(self, value: datetime) -> neo4j.time.DateTime:
        if not isinstance(value, datetime):
            raise ValueError("datetime object expected, got {0}.".format(type(value)))
        return neo4j.time.DateTime.from_native(value)


class JSONProperty(Property):
    """
    Store a data structure as a JSON string.

    The structure will be inflated when a node is retrieved.
    """

    def __init__(self, ensure_ascii: bool = True, *args: Any, **kwargs: Any):
        self.ensure_ascii = ensure_ascii
        super(JSONProperty, self).__init__(*args, **kwargs)

    @validator
    def inflate(self, value: Any) -> Any:
        return json.loads(value)

    @validator
    def deflate(self, value: Any) -> str:
        return json.dumps(value, ensure_ascii=self.ensure_ascii)


class AliasProperty(property, Property):
    """
    Alias another existing property
    """

    def __init__(self, to: str):
        """
        Create new alias

        :param to: name of property aliasing
        :type: str
        """
        super().__init__()
        self.target = to
        self.required = False
        self.has_default = False

    def aliased_to(self) -> str:
        return self.target

    def __get__(self, obj: Any, _type: Optional[Any] = None) -> Property:
        return getattr(obj, self.aliased_to()) if obj else self

    def __set__(self, obj: Any, value: Property) -> None:
        setattr(obj, self.aliased_to(), value)

    @property
    def index(self) -> bool:
        return getattr(self.owner, self.aliased_to()).index

    @index.setter
    def index(self, value: bool) -> None:
        raise AttributeError("Cannot set read-only property 'index'")

    @property
    def unique_index(self) -> bool:
        return getattr(self.owner, self.aliased_to()).unique_index

    @unique_index.setter
    def unique_index(self, value: bool) -> None:
        raise AttributeError("Cannot set read-only property 'unique_index'")


class UniqueIdProperty(Property):
    """
    A unique identifier, a randomly generated uid (uuid4) with a unique index
    """

    def __init__(self, **kwargs: Any):
        for item in ["required", "unique_index", "index", "default"]:
            if item in kwargs:
                raise ValueError(
                    f"{item} argument ignored by {self.__class__.__name__}"
                )

        kwargs["unique_index"] = True
        kwargs["default"] = lambda: uuid.uuid4().hex
        super().__init__(**kwargs)

    @validator
    def inflate(self, value: Any) -> str:
        return str(value)

    @validator
    def deflate(self, value: Any) -> str:
        return str(value)
